#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

          Checklist items for ensuring optimisation is performed as expected.


verify loss @ init. Verify that your loss starts at the correct loss value. E.g. if you initialize your final layer correctly you should measure -log(1/n_classes) on a softmax at initialization. The same default values can be derived for L2 regression, Huber losses, etc.

init well. Initialize the final layer weights correctly. E.g. if you are regressing some values that have a mean of 50 then initialize the final bias to 50. If you have an imbalanced dataset of a ratio 1:10 of positives:negatives, set the bias on your logits such that your network predicts probability of 0.1 at initialization. Setting these correctly will speed up convergence and eliminate “hockey stick” loss curves where in the first few iteration your network is basically just learning the bias.

overfit one batch. Overfit a single batch of only a few examples (e.g. as little as two). To do so we increase the capacity of our model (e.g. add layers or filters) and verify that we can reach the lowest achievable loss (e.g. zero). I also like to visualize in the same plot both the label and the prediction and ensure that they end up aligning perfectly once we reach the minimum loss. If they do not, there is a bug somewhere and we cannot continue to the next stage.

verify decreasing training loss. At this stage you will hopefully be underfitting on your dataset because you’re working with a toy model. Try to increase its capacity just a bit. Did your training loss go down as it should?

visualize just before the net. The unambiguously correct place to visualize your data is immediately before your y_hat = model(x) (or sess.run in tf). That is - you want to visualize exactly what goes into your network, decoding that raw tensor of data and labels into visualizations. This is the only “source of truth”. I can’t count the number of times this has saved me and revealed problems in data preprocessing and augmentation.

use backprop to chart dependencies. Your deep learning code will often contain complicated, vectorized, and broadcasted operations. A relatively common bug I’ve come across a few times is that people get this wrong (e.g. they use view instead of transpose/permute somewhere) and inadvertently mix information across the batch dimension. It is a depressing fact that your network will typically still train okay because it will learn to ignore data from the other examples. One way to debug this (and other related problems) is to set the loss to be something trivial like the sum of all outputs of example i, run the backward pass all the way to the input, and ensure that you get a non-zero gradient only on the i-th input. The same strategy can be used to e.g. ensure that your autoregressive model at time t only depends on 1..t-1. More generally, gradients give you information about what depends on what in your network, which can be useful for debugging.

#TODO: NOT DONE, FINISH!

           Created on 07/07/2020
           """

from random import random

import torch
from draugr.torch_utilities.tensors import to_tensor

__all__ = ["overfit_single_batch"]


# __all__ = ['init_softmax_loss','overfit_single_batch']


def init_softmax_loss():
    """
  #TODO: NOT DONE, FINISH!
  """
    batch_Size = 16
    input_f = 4
    n_classes = 10

    model = torch.nn.Sequential(
        torch.nn.Linear(input_f, 20),
        torch.nn.ReLU(),
        torch.nn.Linear(20, n_classes),
        torch.nn.LogSoftmax(-1),
    )

    for p in model.parameters():
        torch.nn.init.constant_(p, 1)

    input = to_tensor([range(input_f) for _ in range(batch_Size)])

    print(input)  # Visualise input just before forward
    out = model(input)
    print(out)

    target = to_tensor(
        [int(n_classes * random()) for _ in range(batch_Size)], dtype=torch.long
    )
    # loss = torch.nn.MSELoss()(out, target)
    loss = torch.nn.NLLLoss()(out, target)

    expected_loss = -torch.log(to_tensor(1 / n_classes))
    assert expected_loss - loss < 0.1, f"{expected_loss}!={loss}"


def overfit_single_batch():
    """
  #TODO: NOT DONE, FINISH!
  :return:
  """
    input_f = 4
    n_classes = 10

    model = torch.nn.Sequential(
        torch.nn.Linear(input_f, 20),
        torch.nn.ReLU(),
        torch.nn.Linear(20, n_classes),
        torch.nn.LogSoftmax(-1),
    )

    input = to_tensor([range(input_f)])

    print(input)  # Visualise input just before forward
    out = model(input)
    print(out)

    target = torch.zeros(n_classes)
    target[0] = 1
    # loss = torch.nn.MSELoss()(out, target)
    loss = torch.nn.CrossEntropyLoss()(out, target)

    expected_loss = -torch.log(to_tensor(1 / n_classes))
    assert expected_loss == loss, f"{expected_loss}!={loss}"


if __name__ == "__main__":
    init_softmax_loss()
    # overfit_single_batch()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

import math
from typing import Tuple, Union

from draugr import replicate

__all__ = ["max_pool2d_hw_shape"]


def max_pool2d_hw_shape(
    h_w: Union[int, Tuple[int, int]],
    pool_size: Union[int, Tuple[int, int]] = 2,
    stride: Union[int, Tuple[int, int]] = None,
    pad: Union[int, Tuple[int, int]] = 0,
    dilation: Union[int, Tuple[int, int]] = 1,
) -> Tuple[int, int]:
    r"""

  .. math::
  H_{out} = \left\lfloor\frac{H_{in} + 2 * \text{padding[0]} - \text{dilation[0]}
        \times (\text{kernel\_size[0]} - 1) - 1}{\text{stride[0]}} + 1\right\rfloor

  .. math::
  W_{out} = \left\lfloor\frac{W_{in} + 2 * \text{padding[1]} - \text{dilation[1]}
        \times (\text{kernel\_size[1]} - 1) - 1}{\text{stride[1]}} + 1\right\rfloor

  :param pool_size:
  :type pool_size:
  :param h_w:
  :type h_w:
  :param stride:
  :type stride:
  :param pad:
  :type pad:
  :param dilation:
  :type dilation:
  :return:
  :rtype:"""

    if stride is None:
        stride = pool_size
    (h_w, pool_size, stride, pad, dilation) = (
        replicate(h_w),
        replicate(pool_size),
        replicate(stride),
        replicate(pad),
        replicate(dilation),
    )
    pad = (replicate(pad[0]), replicate(pad[1]))

    h = math.floor(
        (h_w[0] + sum(pad[0]) - dilation[0] * (pool_size[0] - 1) - 1) / stride[0] + 1
    )
    w = math.floor(
        (h_w[1] + sum(pad[1]) - dilation[1] * (pool_size[1] - 1) - 1) / stride[1] + 1
    )

    return h, w


if __name__ == "__main__":
    print(max_pool2d_hw_shape(12, 2))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 19/03/2020
           """

from typing import Sequence, Tuple, Union

import cv2
import numpy
from PIL import Image
from draugr.opencv_utilities.bounding_boxes.colors import compute_color_for_labels
from draugr.python_utilities.colors import RGB

__all__ = ["find_contours", "draw_masks"]


def find_contours(*args, **kwargs) -> Tuple:
    """
  Wraps cv2.findContours to maintain compatibility between versions 3 and 4
  Returns:
  contours, hierarchy"""
    if cv2.__version__.startswith("4"):
        contours, hierarchy = cv2.findContours(*args, **kwargs)
    elif cv2.__version__.startswith("3"):
        _, contours, hierarchy = cv2.findContours(*args, **kwargs)
    else:
        raise AssertionError("cv2 must be either version 3 or 4 to call this method")
    return contours, hierarchy


def draw_masks(
    image: Union[Image.Image, numpy.ndarray],
    masks: Union[Image.Image, numpy.ndarray],
    labels: Sequence = None,
    border: bool = True,
    border_width: float = 2,
    border_color: Tuple = RGB(255, 255, 255),
    alpha: float = 0.5,
    color: Tuple = None,
) -> numpy.ndarray:
    """
  Args:
  image: numpy array image, shape should be (height, width, channel)
  masks: (N, 1, Height, Width)
  labels: mask label
  border: draw border on mask
  border_width: border width
  border_color: border color
  alpha: mask alpha
  color: mask color
  Returns:
  numpy.ndarray"""
    if isinstance(image, Image.Image):
        image = numpy.array(image)
    assert isinstance(image, numpy.ndarray)
    if isinstance(masks, Image.Image):
        masks = numpy.array(masks)
    assert isinstance(masks, numpy.ndarray)
    for i, mask in enumerate(masks):
        mask = mask.squeeze()[..., None].astype(numpy.bool)

        label = labels[i] if labels is not None else 1
        _color = compute_color_for_labels(label) if color is None else tuple(color)

        image = numpy.where(
            mask, mask * numpy.array(_color) * alpha + image * (1 - alpha), image
        )
        if border:
            contours, hierarchy = find_contours(
                mask.astype(numpy.uint8), cv2.RETR_TREE, cv2.CHAIN_APPROX_SIMPLE
            )
            image = cv2.drawContours(
                image,
                contours,
                -1,
                border_color,
                thickness=border_width,
                lineType=cv2.LINE_AA,
            )
    return image.astype(numpy.uint8)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/02/2020
           """
__all__ = ["flatten_tn_dim", "flatten_keep_batch", "safe_concat"]

import torch


def flatten_tn_dim(_tensor: torch.tensor) -> torch.tensor:
    """

:param _tensor:
:return:"""
    t, n, *r = _tensor.size()
    return _tensor.reshape(t * n, *r)


def flatten_keep_batch(t: torch.Tensor) -> torch.Tensor:
    """

:param t:
:return:
"""
    return t.reshape(t.shape[0], -1)


def safe_concat(arr: torch.Tensor, el: torch.Tensor, dim: int = 0) -> torch.Tensor:
    """

:param arr:
:param el:
:param dim:
:return:
"""
    if arr is None:
        return el
    return torch.cat((arr, el), dim=dim)


if __name__ == "__main__":

    def a():
        shape = (2, 3, 4, 5)
        from warg import prod

        t = torch.reshape(torch.arange(0, prod(shape)), shape)

        f = flatten_tn_dim(t)
        tf = t.flatten(0, 1)
        print(t, f, tf)
        print(t.shape, f.shape, tf.shape)

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/02/2020
           """

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .info import *
from .mixins import *
from .normalise import *
from .reshaping import *
from .tensor_container import *
from .to_scalar import *
from .to_tensor import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 10/06/2020
           """

import torch
from torch import jit

from warg import AlsoDecorator

__all__ = ["TorchJitSession", "TorchIgnoreJitSession"]


class TorchIgnoreJitSession(AlsoDecorator):
    """
  # Disable torch jit tracing
  """

    def __init__(self, no_side_effect: bool = True):
        self._no_side_effect = no_side_effect
        if no_side_effect:
            self.prev_state = jit._enabled

    def __enter__(self):
        jit._enabled = False
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            jit._enabled = self.prev_state
        else:
            jit._enabled = True


class TorchJitSession(AlsoDecorator):
    """
  # Disable torch jit tracing
  """

    def __init__(self, enabled=False, no_side_effect: bool = True):
        self._no_side_effect = no_side_effect
        self._effect = enabled
        if no_side_effect:
            self.prev_state = jit._enabled

    def __enter__(self):
        jit._enabled = self._effect
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            jit._enabled = self.prev_state
        else:
            jit._enabled = True


if __name__ == "__main__":

    def a():
        @torch.jit.script
        def scripted_fn(x: torch.Tensor):

            for i in range(12):
                x = x + x

            return x

        def fn(x):
            x = torch.neg(x)
            # import pdb
            # pdb.set_trace()
            return scripted_fn(x)

        traced_fn = torch.jit.trace(fn, (torch.rand(4, 5),))
        traced_fn(torch.rand(3, 4))

        print(type(traced_fn))  # torch.jit.ScriptFuntcion

        if isinstance(traced_fn, torch.jit.ScriptFunction):
            # See the compiled graph as Python code
            print(traced_fn.code)

    with TorchIgnoreJitSession():
        a()

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20/03/2020
           """

__all__ = ["compute_color_for_labels"]

from typing import Tuple
from warg import Triple


def compute_color_for_labels(
    label: int, palette: Triple = (2 ** 11 - 1, 2 ** 15 - 1, 2 ** 20 - 1)
) -> Tuple:
    """
  Simple function that adds fixed color depending on the class"""
    return (
        *[int(((label > 0) * p * (label ** 2 - label + 1)) % 255) for p in palette],
    )


if __name__ == "__main__":
    for i in range(9):
        print(compute_color_for_labels(i))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from copy import deepcopy
from typing import Any, Generator, Iterable, Iterator

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/10/2019
           """

__all__ = ["unzip", "unzipper"]


def unzip(iterable: Iterable) -> Iterable:
    return zip(*iterable)


def unzipper(iterable: Iterable[Iterable]) -> Iterable:
    """
  Unzips an iterable of an iterable

  Be carefully has undefined and expected behaviour

  :param iterable:
  :return:"""

    def check_next_iter(iterable: Any) -> Any:
        if isinstance(iterable, Iterable):
            try:
                a = next(iter(iterable))
                if isinstance(a, Iterable):
                    return a
            except StopIteration:
                pass

    if isinstance(iterable, Iterable):
        check_a = check_next_iter(check_next_iter(deepcopy(iterable)))
        if check_next_iter(check_a):
            for a in iterable:
                yield unzipper(a)
        elif check_a:
            for a in iterable:
                yield unzip(a)
        else:
            for i in iterable:
                yield i
    return


if __name__ == "__main__":

    def recursive_eval(node: Any):
        if isinstance(node, (Iterable, Generator, Iterator)):
            gather = []
            for i in node:
                gather.append(recursive_eval(i))
            return gather
        return node

    def aasda():
        r = range(4)

        print(0)

        a = [[[*r] for _ in r] for _ in r]
        print(a)

        print(1)

        for _, assd in zip(r, unzipper(a)):
            print()
            print(recursive_eval(assd))
            print()

        for _, (a, *_) in zip(r, unzipper(a)):
            print()
            print(recursive_eval(a))
            print()

        print(2)

    def skad23():
        print(0)
        zippy_once = zip(range(6), range(3))
        dsadsa = list(deepcopy(zippy_once))
        zippy_twice = zip(dsadsa, dsadsa)
        zippy_twice_copy = deepcopy(zippy_twice)
        asds = list(deepcopy(zippy_twice_copy))
        zippy_trice = zip(asds, asds)
        zippy_trice_copy = deepcopy(zippy_trice)

        print(1)

        for aa in zippy_twice:
            print(recursive_eval(aa))

        print(2)

        for a1 in unzip(zippy_twice_copy):
            print(recursive_eval(a1))

        print(3)

        for a1 in unzip(zippy_once):
            print(recursive_eval(a1))

        print(4)

        for a1 in zippy_trice:
            print(recursive_eval(a1))

        print(5)

        for a1 in unzip(zippy_trice_copy):
            print(recursive_eval(a1))

        print(6)

    def skad():
        print(0)
        zippy_once = zip(zip(range(6), range(3)))
        zippy_once_copy = deepcopy(zippy_once)
        dsadsa = list(deepcopy(zippy_once))
        zippy_twice = zip(dsadsa, dsadsa)
        zippy_twice_copy = deepcopy(zippy_twice)
        asds = list(deepcopy(zippy_twice_copy))
        zippy_trice = zip(asds, asds)
        zippy_trice_copy = deepcopy(zippy_trice)
        asds2323 = list(deepcopy(zippy_trice_copy))
        zippy_quad = zip(asds2323, asds2323)
        zippy_quad_copy = deepcopy(zippy_quad)

        print(1)

        for aa in zippy_twice:
            print(recursive_eval(aa))

        print(2)

        for a1 in unzipper(zippy_twice_copy):
            print(recursive_eval(a1))

        print(3)

        for a1 in zippy_once_copy:
            print(recursive_eval(a1))

        print(4)

        for a1 in unzipper(zippy_once):
            print(recursive_eval(a1))

        print(5)

        for a1 in zippy_trice:
            print(recursive_eval(a1))

        print(6)

        for a1 in unzipper(zippy_trice_copy):
            print(recursive_eval(a1))

        print(7)

        for a1 in zippy_quad:
            print(recursive_eval(a1))

        print(8)

        for a1 in unzipper(zippy_quad_copy):
            print(recursive_eval(a1))

        print(9)

    aasda()

    print()
    print("asafasdw")
    print()

    skad()
    # skad23()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 08-12-2020
           """

__all__ = ["default_worker_init_fn", "global_pin_memory"]

import random
from typing import Union

import numpy
import torch
from draugr.torch_utilities.system.device import global_torch_device


def default_worker_init_fn() -> None:
    worker_seed = torch.initial_seed()
    torch.random.seed(worker_seed)
    random.seed(worker_seed)
    numpy.random.seed(worker_seed)


def global_pin_memory(
    num_workers: int,
    preference: Union[torch.device, bool, str] = True,
    update_num_thread_for_pinned_loader: bool = True,
) -> bool:
    """

#Some weird behaviour of when copying to pinned memory with more workers observed
  @param num_workers:
  @param preference:
  @param update_num_thread_for_pinned_loader:
  @return:


"""
    if isinstance(preference, (torch.device, str)):
        if isinstance(preference, torch.device):
            res = True if "cuda" in torch.device.type else False
        else:
            if "cuda" in preference:
                res = True
            elif preference == "cpu":
                res = False
            else:
                raise NotImplemented
    else:
        res = preference if "cuda" in global_torch_device().type else False

    if num_workers > 1:
        res = False

    if update_num_thread_for_pinned_loader:
        if res and torch.get_num_threads() != 1:
            torch.set_num_threads(1)

    return res


if __name__ == "__main__":

    def main():
        print(global_pin_memory(0))
        print(global_pin_memory(1))
        print(global_pin_memory(2))
        print()
        print(global_pin_memory(0, "cpu"))
        print(global_pin_memory(1, "cpu"))
        print(global_pin_memory(1, "cuda"))
        print()
        print(global_pin_memory(0, "cuda:0"))
        print(global_pin_memory(1, "cuda:0"))

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from draugr.writers.writer import Writer

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""
__all__ = ["ConsoleWriter"]


class ConsoleWriter(Writer):
    """"""

    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        pass

    def _open(self):
        return self

    def _scalar(self, tag: str, value: float, step: int):
        print(f"{step} [{tag}] {value}")


if __name__ == "__main__":

    with ConsoleWriter() as w:
        for i in range(10):
            w.scalar("lol", i)

    print()
    with ConsoleWriter(interval=0) as w:
        for i in range(10):
            w.scalar("lol", i)
    print()
    with ConsoleWriter(interval=-1) as w:
        for i in range(10):
            w.scalar("lol", i)
    print()
    with ConsoleWriter(interval=None) as w:
        for i in range(10):
            w.scalar("lol", i)
    print()
    with ConsoleWriter(interval=2) as w:
        for i in range(10):
            w.scalar("lol", i)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 26-01-2021
           """

from .pandas_utilties import *
from .seaborn_utilities import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04/12/2019
           """

from draugr.torch_utilities import RandomDataset


def test_random_dataset():
    s = (5, 5)
    assert RandomDataset(s, 10)[0].shape == (5, 5)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-09-2020
           """

from typing import Any

import matlab
import numpy

__all__ = ["ndarray_to_matlab", "matlab_to_ndarray"]


def ndarray_to_matlab(x: numpy.ndarray) -> Any:
    """

  :param x:
  :return:
  """
    return matlab.double(x.tolist())


def matlab_to_ndarray(x: Any) -> numpy.ndarray:
    """

  :param x:
  :return:
  """
    return numpy.asarray(x)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

from typing import Tuple, Union

import numpy
from PIL import Image

__all__ = [
    "pil_merge_images",
    "pil_img_to_np_array",
    "np_array_to_pil_img",
    "pil_image_to_byte_array",
    "byte_array_to_pil_image",
]

import io

from PIL import Image


def pil_image_to_byte_array(image: Image.Image, *, coding: str = "PNG") -> bytes:
    """
  PNG encoded by default
  @param coding:
  @param image:
  @return:
  """
    buffer = io.BytesIO()
    image.save(buffer, coding)
    return buffer.getvalue()


def byte_array_to_pil_image(byte_array: bytes) -> Image.Image:
    """

  @param byte_array:
  @return:
  """
    return Image.open(io.BytesIO(byte_array))


def pil_img_to_np_array(
    data_path: Union[str, Image.Image],
    *,
    desired_size: Tuple[int, int] = None,
    expand: int = False
) -> numpy.ndarray:
    """
  Util function for loading RGB image into a numpy array.

  Returns array of shape (1, H, W, C)."""
    img = Image.open(data_path)
    img = img.convert("RGB")
    if desired_size:
        img = img.resize((desired_size[1], desired_size[0]))
    x = numpy.asarray(img, dtype="float32")
    if expand:
        x = numpy.expand_dims(x, axis=0)
    return x / 255.0


def np_array_to_pil_img(x: numpy.ndarray) -> Image.Image:
    """
  Util function for converting anumpy array to a PIL img.

  Returns PIL RGB img."""
    x = numpy.asarray(x)
    x = x + max(-numpy.min(x), 0)
    x_max = numpy.max(x)
    if x_max != 0:
        x /= x_max
    return Image.fromarray((x * 255.0).astype("uint8"), "RGB")


def pil_merge_images(image1: Image.Image, image2: Image.Image) -> Image.Image:
    """Merge two images into one, displayed side by side."""
    (width1, height1) = image1.size
    (width2, height2) = image2.size

    result_width = width1 + width2
    result_height = max(height1, height2)

    result = Image.new("RGB", (result_width, result_height))
    result.paste(im=image1, box=(0, 0))
    result.paste(im=image2, box=(width1, 0))
    return result



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 02-12-2020
           """

__all__ = ["get_model_hash"]

import torch
import hashlib


def get_model_hash(model: torch.nn.Module) -> str:
    model_repr = "".join([str(a) for a in model.named_children()])
    # print(model_repr)
    model_hash = hashlib.md5(model_repr.encode("utf-8")).hexdigest()
    return model_hash



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 05/07/2020
           """

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .bar_writer_mixin import *
from .embed_writer_mixin import *
from .figure_writer_mixin import *
from .graph_writer_mixin import *
from .histogram_writer_mixin import *
from .image_writer_mixin import *
from .line_writer_mixin import *
from .spectrogram_writer_mixin import *
from .precision_recall_writer_mixin import *
from .audio_writer_mixin import *
from .instantiation_writer_mixin import *



from bisect import bisect_right

__all__ = ["WarmupMultiStepLR"]

from torch.optim.lr_scheduler import _LRScheduler


class WarmupMultiStepLR(_LRScheduler):
    """"""

    def __init__(
        self,
        optimiser,
        milestones,
        gamma=0.1,
        warmup_factor=1.0 / 3,
        warmup_iters=500,
        last_epoch=-1,
    ):
        if not list(milestones) == sorted(milestones):
            raise ValueError(
                "Milestones should be a list of" " increasing integers. Got {}",
                milestones,
            )

        self.milestones = milestones
        self.gamma = gamma
        self.warmup_factor = warmup_factor
        self.warmup_iters = warmup_iters
        super().__init__(optimiser, last_epoch)

    def get_lr(self):
        """

    :return:
    :rtype:"""
        warmup_factor = 1
        if self.last_epoch < self.warmup_iters:
            alpha = float(self.last_epoch) / self.warmup_iters
            warmup_factor = self.warmup_factor * (1 - alpha) + alpha
        return [
            base_lr
            * warmup_factor
            * self.gamma ** bisect_right(self.milestones, self.last_epoch)
            for base_lr in self.base_lrs
        ]



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 16/02/2020
           """

import math

__all__ = ["lambda_accumulator", "mean_accumulator", "total_accumulator"]


def lambda_accumulator(start_value=None, lambd=0.99):
    """

  :param start_value:
  :type start_value:
  :param lambd:
  :type lambd:
  :return:
  :rtype:"""
    assert 0 <= lambd <= 1

    def lambda_accumulator_(n=start_value):
        """

    :param n:
    :type n:"""
        while True:
            new_n = yield n
            if new_n is not None:
                if n is not None:
                    n = n * lambd + new_n * (1 - lambd)
                else:
                    n = new_n

    acc = lambda_accumulator_()
    acc.send(None)

    return acc


def mean_accumulator(start_value=None):
    """

  :param start_value:
  :type start_value:
  :return:
  :rtype:"""

    def mean_accumulator_(n=start_value):
        """

    :param n:
    :type n:"""
        if n is not None:
            num = 1
        else:
            num = 0
        while True:
            new_n = yield n
            if new_n is not None:
                num += 1
                if n is not None:
                    n = n + (new_n - n) / num
                else:
                    n = new_n

    acc = mean_accumulator_()
    acc.send(None)

    return acc


def total_accumulator(start_value=0):
    """

  :param start_value:
  :type start_value:
  :return:
  :rtype:"""

    def total_accumulator_(total=start_value):
        """

    :param total:
    :type total:"""
        while True:
            a = yield total
            if a:
                total += a

    acc = total_accumulator_()
    acc.send(None)

    return acc


if __name__ == "__main__":
    samples = 10000

    def stest_1():
        """"""
        lmbd_acc = lambda_accumulator()

        print(next(lmbd_acc))

        for i in range(samples):
            lmbd_acc.send(math.cos(i))
            if i < 5 or i > samples - 5:
                print(next(lmbd_acc))

    def stest_2():
        """"""
        mean_acc = mean_accumulator()

        print(next(mean_acc))

        for i in range(samples):
            mean_acc.send(math.cos(i))
            if i < 5 or i > samples - 5:
                print(next(mean_acc))

    def stest_3():
        """"""
        mean_acc = mean_accumulator()
        print(next(mean_acc))

        for i in range(samples):
            mean_acc.send(1)
            if i < 5 or i > samples - 5:
                print(next(mean_acc))

    def stest_4():
        """"""
        lmbd_acc = lambda_accumulator()

        print(next(lmbd_acc))

        for i in range(samples):
            lmbd_acc.send(1)
            if i < 5 or i > samples - 5:
                print(next(lmbd_acc))

    def stest_lambda_zero():
        """"""
        lmbd_acc = lambda_accumulator(lambd=0)

        print(next(lmbd_acc))

        for i in range(samples):
            lmbd_acc.send(1)
            if i < 5 or i > samples - 5:
                print(next(lmbd_acc))

    def stest_lambda_zero_input():
        """"""
        lmbd_acc = lambda_accumulator()

        print(next(lmbd_acc))

        for i in range(samples):
            lmbd_acc.send(1)
            if i < 5 or i > samples - 5:
                print(next(lmbd_acc))

    def stest_5():
        """"""
        total = total_accumulator()

        print(next(total))

        for i in range(samples):
            total.send(1)
            if i < 5 or i > samples - 5:
                print(next(total))
        print(next(total))

    def stest_6():
        """"""
        total = total_accumulator()

        print(next(total))

        for i in range(samples):
            total.send(i)
            if i < 5 or i > samples - 5:
                print(next(total))
        print(next(total))

    stest_1()
    stest_2()
    stest_3()
    stest_4()
    stest_5()
    stest_6()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import functools
from typing import Any, Callable

import numpy
import torch

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["torch_seed"]


def torch_seed(s: int = 72163) -> torch.Generator:
    """
  seeding for reproducibility"""
    generator = torch.manual_seed(s)
    if False:  # Disabled for now
        torch.set_deterministic(True)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(s)
        torch.backends.cudnn.deterministic = True
    return generator


class Seed:
    r"""**Seed PyTorch and numpy.**

  This code is based on PyTorch's reproducibility guide: https://pytorch.org/docs/stable/notes/randomness.html
  Can be used as standard seeding procedure, context manager (seed will be changed only within block) or function decorator.

  **Standard seed**::

        torchfunc.Seed(0) # no surprises I guess

  **Used as context manager**::

    with Seed(1):
        ... # your operations

    print(torch.initial_seed()) # Should be back to seed pre block

  **Used as function decorator**::

    @Seed(1) # Seed only within function
    def foo():
        return 42

  **Important:** It's impossible to put original `numpy` seed after context manager
  or decorator, hence it will be set to original PyTorch's seed.

  Parameters
  ----------
  value: int
        Seed value used in np.random_seed and torch.manual_seed. Usually int is provided
  cuda: bool, optional
        Whether to set PyTorch's cuda backend into deterministic mode (setting cudnn.benchmark to `False`
        and cudnn.deterministic to `True`). If `False`, consecutive runs may be slightly different.
        If `True`, automatic autotuning for convolutions layers with consistent input shape will be turned off.
        Default: `False`
  """

    def __init__(self, value: int, cuda: bool = False):
        self.value = value
        self.cuda = cuda

        self.no_side_effect = False
        if self.no_side_effect:
            self._last_seed = torch.initial_seed()
        numpy.random.seed(self.value)
        torch.manual_seed(self.value)
        if False:  # Disabled for now
            torch.set_deterministic(True)
        if self.cuda:
            torch.backends.cudnn.deterministic = True
            torch.backends.cudnn.benchmark = False

    def __enter__(self):
        return self

    def __exit__(self, *_, **__):
        if self.no_side_effect:
            torch.manual_seed(self._last_seed)
            numpy.random.seed(self._last_seed)
        return False

    def __call__(self, function: Callable) -> callable:
        @functools.wraps(function)
        def decorated(*args, **kwargs) -> Any:
            value = function(*args, **kwargs)
            self.__exit__()
            return value

        return decorated


if __name__ == "__main__":

    @Seed(1)  # Seed only within function
    def foo():
        return torch.randint(5, (2, 2))

    def bar():
        with Seed(1):
            return torch.randint(5, (2, 2))

    def buzz():
        Seed(1)
        return torch.randint(5, (2, 2))

    for f in [foo, bar, buzz]:
        print(f())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 30-08-2020
           """

import winreg

from warg import SingletonMeta


class WindowsSettings(metaclass=SingletonMeta):
    """"""

    def __init__(self):
        self.registry = winreg.ConnectRegistry(None, winreg.HKEY_CURRENT_USER)
        self.changes = set()

    def get_dark_mode(self) -> bool:
        """
    True if

    :return:"""
        key = winreg.OpenKey(
            self.registry,
            r"SOFTWARE\Microsoft\Windows\CurrentVersion\Themes\Personalize",
        )
        return False if winreg.QueryValueEx(key, "AppsUseLightTheme")[0] else "black"

    def set_dark_mode(self, value: bool) -> None:
        key = winreg.OpenKey(
            self.registry,
            r"SOFTWARE\Microsoft\Windows\CurrentVersion\Themes\Personalize",
            0,
            winreg.KEY_ALL_ACCESS | winreg.KEY_WOW64_64KEY,
        )
        self.changes.add(key)

        # noinspection PyTypeChecker
        winreg.SetValueEx(
            key,
            "AppsUseLightTheme",
            None,
            winreg.REG_DWORD,
            int(not value),  # SHOULD BE OF TYPE INT!
        )

    def __enter__(self):
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        return True
        # FlushKey
        # for c in self.changes:
        #  pass
        # self.registry

    @property
    def all_settings(self):
        """"""
        return None
        # for i in self.settings.list_properties():  # getting all existing properties
        #  yield i


if __name__ == "__main__":

    def a():
        ws = WindowsSettings()
        dark_mode = ws.get_dark_mode()
        ws.set_dark_mode(not dark_mode)

    a()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["FigureWriterMixin"]

from matplotlib.figure import Figure


class FigureWriterMixin(ABC):
    """
  Writer mixin that provides an interface for 'writing' figures"""

    @abstractmethod
    def figure(self, tag: str, figure: Figure, step: int, **kwargs) -> None:
        """

    :param tag:
    :type tag:
    :param data:
    :type data:
    :param step:
    :type step:
    :param dataformats:
    :type dataformats:
    :param kwargs:
    :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-01-2021
           """

from threading import Thread
from typing import Any

import cv2

__all__ = ["AsyncVideoStream"]


class AsyncVideoStream:
    def __init__(self, src=0, thread_name: str = None, group: Any = None):
        """
    threaded async wrapper around opencv cv2.VideoCapture with alike interface

    @param src:
    @param thread_name:
    """
        self._stream = cv2.VideoCapture(src)
        (self.grabbed, self.frame) = self._stream.read()

        self._stopped = False
        self._thread = Thread(
            target=self.update, name=thread_name, args=(), daemon=True, group=group
        )

    def start(self):
        """
        # start the thread to read frames from the video stream
    @return:
    """

        self._thread.start()
        return self

    def update(self):
        while not self._stopped:  # keep looping infinitely until the thread is stopped
            (
                self.grabbed,
                self.frame,
            ) = self._stream.read()  # otherwise, read the next frame from the stream

    def read(self):
        return self.grabbed, self.frame  # return the frame most recently read

    def stop(self):
        self._stream.release()
        self._stopped = True  # indicate that the thread should be stopped

    # noinspection PyPep8Naming
    def isOpened(self):
        return self._stream.isOpened()

    def __call__(self, *args, **kwargs):
        return self.frame

    def __next__(self):
        return self.frame

    def __iter__(self):
        return self.start()

    def __del__(self):
        self.stop()
        self._thread.join()

    def __enter__(self):
        return self.start()

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.stop()


if __name__ == "__main__":
    for a in AsyncVideoStream():
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"

import torch

__all__ = ["soft_update"]


def soft_update(
    target: torch.nn.Module, source: torch.nn.Module, tau: float
) -> torch.nn.Module:
    assert 0 <= tau <= 1
    for target_param, param in zip(target.parameters(), source.parameters()):
        target_param.data.copy_(target_param.data * (1.0 - tau) + param.data * tau)

    return target



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25/03/2020
           """

from .boxes_2d import *
from .colors import *
from .draw_boxes import *
from .evaluation import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"

import numpy
import visdom

vis = visdom.Visdom()

__all__ = ["plot_episode_stats"]


def plot_episode_stats(stats):
    """

  :param stats:
  :type stats:
  :return:
  :rtype:"""
    vis.line(
        X=numpy.arange(len(stats.signal_mas)),
        Y=numpy.array(stats.signal_mas),
        win="DDPG MEAN REWARD (100 episodes)",
        opts=dict(
            title=("DDPG MEAN REWARD (100 episodes)"),
            ylabel="MEAN REWARD (100 episodes)",
            xlabel="Episode",
        ),
    )  # Plot the mean of last 100 episode rewards over time.

    vis.line(
        X=numpy.cumsum(stats.episode_lengths),
        Y=numpy.arange(len(stats.episode_lengths)),
        win="DDPG Episode per time step",
        opts=dict(
            title=("DDPG Episode per time step"), ylabel="Episode", xlabel="Time Steps"
        ),
    )  # Plot time steps and episode number.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-01-2021
           """

from typing import Sequence

import cv2
from draugr.drawers.drawer import Drawer
from warg import passes_kws_to


class OpencvImageStream(Drawer):
    """"""

    def __init__(self, title: str = "", render: bool = True, **kwargs):

        if not render:
            return

        self.window_id = title
        cv2.namedWindow(self.window_id, cv2.WINDOW_NORMAL)

    def draw(self, data: Sequence):
        """

    :param data:
    :return:"""
        if cv2.waitKey(1) == 27:  # esc to quit
            cv2.destroyWindow(self.window_id)
            raise StopIteration
        else:
            cv2.imshow(self.window_id, data)


if __name__ == "__main__":

    def asdasf():

        from draugr.opencv_utilities import frame_generator, AsyncVideoStream
        from draugr.tqdm_utilities import progress_bar

        with AsyncVideoStream() as vc:
            with OpencvImageStream() as s:
                for i in progress_bar(
                    frame_generator(vc, coder=None), auto_total_generator=False
                ):
                    s.draw(i)

    asdasf()



__all__ = ["NamedTensorTuple"]


class NamedTensorTuple:
    """
  Help class for manage boxes, labels, etc...
  Not inherit dict due to `default_collate` will change dict's subclass to dict."""

    def __init__(self, **kwargs):
        self._data_dict = kwargs

    def __setattr__(self, key, value):
        object.__setattr__(self, key, value)

    def __getitem__(self, key):
        return self._data_dict[key]

    def __iter__(self):
        return self._data_dict.__iter__()

    def __setitem__(self, key, value):
        self._data_dict[key] = value

    def _call(self, name, *args, **kwargs):
        keys = list(self._data_dict.keys())
        for key in keys:
            value = self._data_dict[key]
            if hasattr(value, name):
                self._data_dict[key] = getattr(value, name)(*args, **kwargs)
        return self

    def to(self, *args, **kwargs):
        """

    :param args:
    :type args:
    :param kwargs:
    :type kwargs:
    :return:
    :rtype:"""
        return self._call("to", *args, **kwargs)

    def numpy(self):
        """

    :return:
    :rtype:"""
        return self._call("numpy")

    def __repr__(self):
        return self._data_dict.__repr__()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import datetime
import os
import pathlib
import sys
from typing import Union

import torch
from draugr.torch_utilities.persistence.config import (
    ensure_directory_exist,
    save_config,
)
from torch.nn.modules.module import Module
from warg import passes_kws_to
from warg.decorators.kw_passing import drop_unused_kws

__author__ = "Christian Heider Nielsen"

model_extension = ".model"
config_extension = ".py"

__all__ = [
    "load_model",
    "load_latest_model",
    "save_model_and_configuration",
    "save_model",
    "convert_saved_model_to_cpu",
]


@drop_unused_kws
def load_latest_model(
    *, model_name: str, model_directory: pathlib.Path, raise_on_failure: bool = True
) -> Union[torch.nn.Module, None]:
    """

load model with the lastest time appendix or in this case creation time

:param raise_on_failure:
:param model_directory:
:param model_name:
:return:"""
    model_path = model_directory / model_name
    list_of_files = list(model_path.glob(f"*{model_extension}"))
    if len(list_of_files) == 0:
        msg = (
            f"Found no previous models with extension {model_extension} in {model_path}"
        )
        if raise_on_failure:
            raise FileNotFoundError(msg)
        print(msg)
        return None
    latest_model = max(list_of_files, key=os.path.getctime)  # USES CREATION TIME
    print(f"loading previous model: {latest_model}")

    return torch.load(str(latest_model))


load_model = load_latest_model


@passes_kws_to(save_config)
def save_model_and_configuration(
    *,
    model: Module,
    model_save_path: pathlib.Path,
    config_save_path: pathlib.Path = None,
    loaded_config_file_path: pathlib.Path = None,
    raise_on_existing: bool = False,
) -> None:
    """

:param raise_on_existing:
:param model:
:param model_save_path:
:param config_save_path:
:param loaded_config_file_path:
:return:"""
    if raise_on_existing and model_save_path.exists():
        raise FileExistsError(f"{model_save_path} exists!")
    torch.save(model, str(model_save_path))
    if loaded_config_file_path:
        save_config(config_save_path, loaded_config_file_path)


@drop_unused_kws
@passes_kws_to(save_model_and_configuration)
def save_model(
    model: Module,
    *,
    model_name: str,
    save_directory: pathlib.Path,  # TODO: RENAME to model directory for consistency
    config_file_path: pathlib.Path = None,
    prompt_on_failure: bool = True,
    verbose: bool = False,
) -> None:
    """

save a model with a timestamp appendix to later to loaded

:param prompt_on_failure:
:param verbose:
:param model:
:param save_directory:
:param config_file_path:
:param model_name:
:return:"""
    model_date = datetime.datetime.now()
    # config_name = config_name.replace(".", "_")

    model_time_rep = model_date.strftime("%Y%m%d%H%M%S")
    model_save_path = save_directory / model_name / f"{model_time_rep}{model_extension}"
    config_save_path = (
        save_directory / model_name / f"{model_time_rep}{config_extension}"
    )
    ensure_directory_exist(model_save_path.parent)

    saved = False
    try:
        save_model_and_configuration(
            model=model,
            model_save_path=model_save_path,
            loaded_config_file_path=config_file_path,
            config_save_path=config_save_path,
        )
        saved = True
    except FileNotFoundError as e:
        if prompt_on_failure:
            print(e)
            while not saved:
                file_path = input("Enter another file path: ")
                model_save_path = pathlib.Path(file_path).expanduser().resolve()
                parent = model_save_path.parent
                ensure_directory_exist(parent)
                config_save_path = parent / f"{model_save_path.name}{config_extension}"
                try:
                    save_model_and_configuration(
                        model=model,
                        model_save_path=model_save_path,
                        loaded_config_file_path=config_file_path,
                        config_save_path=config_save_path,
                    )
                    saved = True
                except FileNotFoundError as e:
                    print(e)
                    saved = False
        else:
            raise e

    if verbose:
        if saved:
            print(
                f"Successfully saved model and configuration respectively at {model_save_path} and {config_save_path}"
            )
        else:
            print(f"Was unsuccesful at saving model or configuration")


def convert_saved_model_to_cpu(path: pathlib.Path) -> None:
    """

:param path:
:return:"""
    model = torch.load(path, map_location=lambda storage, loc: storage)
    torch.save(model, f"{path}.cpu{model_extension}")


if __name__ == "__main__":
    convert_saved_model_to_cpu(pathlib.Path(sys.argv[1]))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """

__all__ = ["indent_lines"]

from typing import Any


def indent_lines(
    input_str: Any, indent_spaces_num: int = 2, ignore_single_lines: bool = False
) -> str:
    """

  :param ignore_single_lines:
  :type ignore_single_lines:
  :param input_str:
  :type input_str:
  :param indent_spaces_num:
  :type indent_spaces_num:
  :return:
  :rtype:"""
    if not isinstance(input_str, str):
        input_str = str(input_str)
    s = input_str.split("\n")
    indent_s = indent_spaces_num * " "
    if len(s) == 1:
        if ignore_single_lines:
            return input_str
        else:
            return f"{indent_s}{input_str}"
    first = s.pop(0)
    s = [f"{indent_s}{line}" for line in s]
    s = "\n".join(s)
    s = f"{indent_s}{first}\n{s}"
    return s


if __name__ == "__main__":
    a = "slasc\nsaffasd\n2dasf"
    print(a)
    print(indent_lines(a))



# To add a new cell, type '# %%'
# To add a new markdown cell, type '# %% [markdown]'
# %% [markdown]
# # Bad gradients in PyTorch graph

# %%

import torch
from draugr.torch_utilities import register_bad_grad_hooks

x = torch.randn(10, 10, requires_grad=True)
y = torch.randn(10, 10, requires_grad=True)

z = x / (y * 0)
z = z.sum() * 2
get_dot = register_bad_grad_hooks(z)
z.backward()
dot = get_dot()
# dot.save('tmp.dot') # to get .dot
# dot.render('tmp') # to get SVG
dot  # in Jupyter, you can just render the variable



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 07-01-2021
           """

from math import ceil, log

import numpy
from warg import Number

__all__ = [
    "next_pow_numpy",
    "next_pow_2_numpy",
    "next_pow_10_numpy",
    "next_power_2_numpy",
    "next_power_10_numpy",
]


def next_pow_numpy(x: Number, n: int) -> int:
    """Calculates the next power of n of a number."""

    return int(pow(n, numpy.ceil(numpy.log(x) / numpy.log(n))))


def next_pow_2_numpy(x: Number) -> int:
    """Calculates the next power of 2 of a number."""

    return int(pow(2, numpy.ceil(numpy.log2(x))))


next_power_2_numpy = next_pow_2_numpy


def next_pow_10_numpy(x: Number) -> int:
    """Calculates the next power of 10 of a number."""

    return int(pow(10, numpy.ceil(numpy.log10(x))))


next_power_10_numpy = next_pow_10_numpy

if __name__ == "__main__":
    for i in range(1, 11 + 1):
        print(next_pow_2_numpy(i))
        print(next_pow_numpy(i, 2))
        # print(next_pow(i, 3))
        # print(next_pow_10(i))

    print(next_pow_2_numpy(next_pow_2_numpy(next_pow_2_numpy(2))))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 08-12-2020
           """

from draugr.torch_utilities import global_pin_memory


def test_load_pin_memory_global():
    print(global_pin_memory(0))
    print(global_pin_memory(1))
    print(global_pin_memory(2))
    print()
    print(global_pin_memory(0, "cpu"))
    print(global_pin_memory(1, "cpu"))
    print(global_pin_memory(1, "cuda"))
    print()
    print(print(global_pin_memory(0, "cuda:0")))
    print(print(global_pin_memory(1, "cuda:0")))


if __name__ == "__main__":
    test_load_pin_memory_global()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from draugr.writers.writer import Writer

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["MockWriter"]


class MockWriter(Writer):
    """"""

    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        pass

    def _open(self):
        return self

    def _scalar(self, *args, **kwargs):
        pass


if __name__ == "__main__":
    with MockWriter() as w:
        w.scalar("a", 2)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from abc import ABC, abstractmethod

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """
__all__ = ["PrecisionRecallCurveWriterMixin"]

from typing import Mapping

from warg import drop_unused_kws


class PrecisionRecallCurveWriterMixin(ABC):
    """
  Writer mixin that provides an interface for 'writing' instantiation"""

    @drop_unused_kws
    @abstractmethod
    def precision_recall_curve(
        self,
        tag: str,
        predictions: list,
        truths: list,
        step: int,
        num_thresholds: int = 11,
    ) -> None:
        """

    :param tag:
    :type tag:
    :param data:
    :type data:
    :param step:
    :type step:
    :param dataformats:
    :type dataformats:
    :param kwargs:
    :type kwargs:"""
        raise NotImplementedError



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/04/2020
           """

from typing import Sequence, Tuple

import numpy

__all__ = ["intersect_numpy", "jaccard_overlap_numpy", "remove_null_boxes"]


def intersect_numpy(box_a: Sequence, box_b: Sequence) -> numpy.ndarray:
    """

  :param box_a:
  :type box_a:
  :param box_b:
  :type box_b:
  :return:
  :rtype:"""
    max_xy = numpy.minimum(box_a[:, 2:], box_b[2:])
    min_xy = numpy.maximum(box_a[:, :2], box_b[:2])
    inter = numpy.clip((max_xy - min_xy), a_min=0, a_max=numpy.inf)
    return inter[:, 0] * inter[:, 1]


def jaccard_overlap_numpy(box_a: numpy.ndarray, box_b: numpy.ndarray) -> numpy.ndarray:
    """Compute the jaccard overlap of two sets of boxes.  The jaccard overlap
  is simply the intersection over union of two boxes.
  E.g.:
  A ∩ B / A ∪ B = A ∩ B / (area(A) + area(B) - A ∩ B)
  Args:
  box_a: Multiple bounding boxes, Shape: [num_boxes,4]
  box_b: Single bounding box, Shape: [4]
  Return:
  jaccard overlap: Shape: [box_a.shape[0], box_a.shape[1]]"""
    inter = intersect_numpy(box_a, box_b)
    area_a = (box_a[:, 2] - box_a[:, 0]) * (box_a[:, 3] - box_a[:, 1])  # [A,B]
    area_b = (box_b[2] - box_b[0]) * (box_b[3] - box_b[1])  # [A,B]
    union = area_a + area_b - inter
    return inter / union  # [A,B]


def remove_null_boxes(
    boxes: numpy.ndarray, labels: numpy.ndarray
) -> Tuple[numpy.ndarray, numpy.ndarray]:
    """Removes bounding boxes of W or H equal to 0 and its labels

  Args:
  boxes   (ndarray): NP Array with bounding boxes as lines
                 * BBOX[x1, y1, x2, y2]
  labels  (labels): Corresponding labels with boxes

  Returns:
  ndarray: Valid bounding boxes
  ndarray: Corresponding labels"""
    del_boxes = []
    for idx, box in enumerate(boxes):
        if box[0] == box[2] or box[1] == box[3]:
            del_boxes.append(idx)

    return numpy.delete(boxes, del_boxes, 0), numpy.delete(labels, del_boxes)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from draugr.writers.writer import Writer

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 9/14/19
           """

__all__ = ["VisdomWriter"]

# Visualisation
USE_VISDOM = False
START_VISDOM_SERVER = False
VISDOM_SERVER = "http://localhost"
if not START_VISDOM_SERVER:
    # noinspection PyRedeclaration
    VISDOM_SERVER = "http://visdom.ml"


class VisdomWriter(Writer):
    """"""

    def _scalar(self, tag: str, value: float, step: int):
        pass

    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        pass

    def _open(self):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11/05/2020
           """

import torch

from draugr.torch_utilities import (
    TorchCpuSession,
    TorchCudaSession,
    TorchEvalSession,
    TorchTrainSession,
    global_torch_device,
)


def test_cpu():
    print(
        global_torch_device(override=global_torch_device(cuda_device_preference=True))
    )
    print(global_torch_device())
    with TorchCpuSession():
        print(global_torch_device())
    print(global_torch_device())


def test_nested_model_sessions():
    model = torch.nn.Sequential(torch.nn.Linear(1, 1), torch.nn.Dropout(0.1))
    print(model.training)
    with TorchEvalSession(model):
        print(model.training)
        with TorchTrainSession(model):
            print(model.training)
            with TorchEvalSession(model):
                print(model.training)
                with TorchTrainSession(model):
                    print(model.training)
                    with TorchEvalSession(model):
                        print(model.training)
    print(model.training)


def test_nested_device_sessions():
    print(
        global_torch_device(override=global_torch_device(cuda_device_preference=True))
    )
    print(global_torch_device())
    with TorchCpuSession():
        print(global_torch_device())
        with TorchCudaSession():
            print(global_torch_device())
            with TorchCpuSession():
                print(global_torch_device())
    print(global_torch_device())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""
"""

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

import warnings

warnings.filterwarnings(
    "ignore", category=FutureWarning
)  # Because Tensorflow needs to remind everyone all
# the time that their system is constantly being deprecated..
from .launcher import *
from .tensorboard_pytorch_writer import *

# from .tensorboard_x_writer import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 29/07/2020
           """

__all__ = []

import copy
import math

import torch
from draugr.torch_utilities.optimisation.parameters import (
    named_trainable_parameters,
    normal_init_weights,
    trainable_parameters,
)
from draugr.torch_utilities.sessions import TorchEvalSession
from warg import ContextWrapper


def loss_grad_check(
    model: torch.nn.Module,
    loss_fn: callable,
    input: torch.Tensor,
    target: torch.Tensor,
    epsilon: float = 1e-6,
    error_tolerance: float = 1e-5,
) -> None:
    """
  two sided gradient numerical approximation
  DOES not work, please refer to torch/autograd/gradcheck.py

  :param input:
  :type input:
  :param target:
  :type target:
  :param error_tolerance:
  :type error_tolerance:
  :param model:
  :type model:
  :param loss_fn:
  :type loss_fn:
  :param epsilon:
  :type epsilon:
  :return:
  :rtype:"""
    assert epsilon > 0.0
    c_model = copy.deepcopy(model)

    loss = loss_fn(model(input), target)
    loss.backward()
    with ContextWrapper(torch.no_gra, True):
        with TorchEvalSession(model):
            for (n, c_p), p in zip(
                named_trainable_parameters(c_model).items(), trainable_parameters(model)
            ):
                for i, c_p_o in enumerate(c_p):
                    a = c_p_o.size()
                    if len(a) > 0:
                        for j in range(a[0]):
                            cp_orig = c_p.data.clone()

                            c_p[i][j] += epsilon  # positive
                            loss_p = loss_fn(
                                c_model(input.clone()), target.clone()
                            ).clone()

                            c_p.data = cp_orig

                            c_p[i][j] -= epsilon  # negative
                            loss_n = loss_fn(
                                c_model(input.clone()), target.clone()
                            ).clone()

                            c_p.data = cp_orig

                            if (
                                True
                            ):  # TODO: make check based on the entire set of parameters at once
                                grad_approx = (loss_p - loss_n) / (2 * epsilon)

                                denom = math.sqrt(grad_approx ** 2) + math.sqrt(
                                    p.grad[i][j] ** 2
                                )
                                if denom > 0:
                                    deviance = (
                                        math.sqrt((grad_approx - p.grad[i][j]) ** 2)
                                        / denom
                                    )
                                    # assert torch.sign(grad_approx) == torch.sign(p.grad[i][j]), f'apprx: {grad_approx}, analytical {p.grad[i][j]}'
                                    assert (
                                        deviance <= error_tolerance
                                    ), f"Numerical gradient approximation of parameter {n} deviates larger than tolerance {error_tolerance}, deviance: {deviance}, approx:{grad_approx, loss_p, loss_n}, p.grad[i][j]:{p.grad[i][j]}"
                                else:
                                    pass
                                    # print(grad_approx,denom)


if __name__ == "__main__":

    def stest_return_duplicate():
        from torch.autograd import Function, gradcheck, gradgradcheck

        class DoubleDuplicate(Function):
            @staticmethod
            def forward(ctx, x):
                output = x * 2
                return output, output

            @staticmethod
            def backward(ctx, grad1, grad2):
                return grad1 * 2 + grad2 * 2

        def fn(x):
            a, b = DoubleDuplicate.apply(x)
            return a + b

        x = torch.randn(5, 5, requires_grad=True, dtype=torch.double)
        gradcheck(fn, [x], eps=1e-6)
        gradgradcheck(fn, [x])

    def a():
        #    from torch.testing import _get_default_tolerance

        input = torch.randn(5, 5, requires_grad=True, dtype=torch.double)
        target = torch.randn(5, 5, requires_grad=False, dtype=torch.double)
        model = torch.nn.Sequential(
            torch.nn.Linear(5, 50),
            torch.nn.ReLU(),
            torch.nn.Linear(50, 50),
            torch.nn.Linear(50, 5),
        ).double()
        normal_init_weights(model)
        criterion = torch.nn.MSELoss()
        # _get_default_tolerance(input)
        loss_grad_check(model, criterion, input, target)

    # a()
    stest_return_duplicate()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/07/2020
           """

__all__ = ["recursive_flatten"]

from typing import Sequence


def recursive_flatten(seq: Sequence) -> Sequence:
    if not seq:  # is empty Sequence
        return seq
    if isinstance(seq[0], Sequence):
        return (*recursive_flatten(seq[0]), *recursive_flatten(seq[1:]))
    return (*seq[:1], *recursive_flatten(seq[1:]))


if __name__ == "__main__":

    print(recursive_flatten((((2,), 2), (2,), 2)))
    print(recursive_flatten((([[None]], 2), (2,), 2)))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = ""
__all__ = ["launch_tensorboard"]

from pathlib import Path


def launch_tensorboard(log_dir: Path, port: int = 6006) -> str:
    """

  :param log_dir:
  :type log_dir:
  :param port:
  :type port:
  :return:
  :rtype:"""
    from tensorboard import program

    tb = program.TensorBoard()
    # tb.configure(argv=['', '--logdir', log_dir, '--port', port])
    tb.configure(logdir=str(log_dir), port=port)
    return tb.launch()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/02/2020
           """

__all__ = ["frozen_parameters", "freeze_parameters"]

from contextlib import contextmanager
from itertools import tee
from typing import Iterator

from torch.nn import Parameter


def freeze_parameters(params: Iterator[Parameter], value: bool = None) -> None:
    """

  :param params:
  :param value:
  :return:"""
    if isinstance(value, bool):
        for p in params:
            p.requires_grad = not value
    else:
        for p in params:
            p.requires_grad = not p.requires_grad


@contextmanager
def frozen_parameters(params: Iterator[Parameter], enabled=True) -> None:
    """

  :param enabled:
  :type enabled:
  :param params:
  :return:"""
    params_1, params_2 = tee(params)
    if enabled:
        freeze_parameters(params_1, True)
    yield True
    if enabled:
        freeze_parameters(params_2, False)


if __name__ == "__main__":
    from torch import nn

    def asd21312a():
        a = nn.Linear(10, 5)
        print(a.weight.requires_grad)
        with frozen_parameters(a.parameters()):
            print(a.weight.requires_grad)
        print(a.weight.requires_grad)

    def afsda32():
        a = nn.Linear(10, 5)

        print(a.weight.requires_grad)
        with frozen_parameters(a.parameters()):
            print(a.weight.requires_grad)
        print(a.weight.requires_grad)

    def afsda12332_toogle():
        a = nn.Linear(10, 5)

        print(a.weight.requires_grad)
        freeze_parameters(a.parameters())
        print(a.weight.requires_grad)
        freeze_parameters(a.parameters())
        print(a.weight.requires_grad)

    def afsda12332_explicit():
        a = nn.Linear(10, 5)

        print(a.weight.requires_grad)
        freeze_parameters(a.parameters(), True)
        print(a.weight.requires_grad)
        freeze_parameters(a.parameters(), False)
        print(a.weight.requires_grad)

    def seq_no_context():
        a = nn.Sequential(nn.Linear(10, 5), nn.Linear(5, 5))

        print(next(a.parameters()).requires_grad)
        freeze_parameters(a.parameters(), True)
        print(next(a.parameters()).requires_grad)
        freeze_parameters(a.parameters(), False)
        print(next(a.parameters()).requires_grad)

    def seq_context():
        a = nn.Sequential(nn.Linear(10, 5), nn.Linear(5, 5))

        print(next(a.parameters()).requires_grad)
        freeze_parameters(a.parameters(), True)
        print(next(a.parameters()).requires_grad)
        freeze_parameters(a.parameters(), False)
        print(next(a.parameters()).requires_grad)

    print("\n")
    asd21312a()
    print("\n")
    afsda32()
    print("\n")
    afsda12332_toogle()
    print("\n")
    afsda12332_explicit()
    print("\n")
    seq_no_context()
    print("\n")
    seq_context()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17-09-2020
           """

from .. import INCLUDE_PROJECT_READMES

if INCLUDE_PROJECT_READMES:
    import pathlib

    with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
        __doc__ += this_init_file.read()

from .conversion import *
from .singleton_engine import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


from collections import Sized
from typing import Any, Callable, Generator, Iterable, List

from pathos.helpers import cpu_count
from pathos.multiprocessing import ProcessPool as Pool
from tqdm.auto import tqdm

__author__ = "Christian Heider Nielsen"
__all__ = [
    "parallel_map",
    "parallel_imap",
    "parallel_umap",
    "parallel_uimap",
    "sequential_map",
    "sequential_imap",
]
__doc__ = (
    r"""Map functions with tqdm progress bars for parallel and sequential processing."""
)


def _sequential(
    function: Callable, *iterables: Iterable, func_kws=None, **kwargs: Any
) -> Generator:
    """Returns a generator for a sequential map with a progress bar.

  Arguments:
      function(Callable): The function to apply to each element of the given Iterables.
      iterables(Tuple[Iterable]): One or more Iterables containing the data to be mapped.

  Returns:
      A generator which will apply the function to each element of the given Iterables
      sequentially in order with a progress bar.
  """

    # Determine length of tqdm (equal to length of shortest iterable)
    length = min(len(iterable) for iterable in iterables if isinstance(iterable, Sized))

    # Create sequential generator
    for item in tqdm(map(function, *iterables, **func_kws), total=length, **kwargs):
        yield item


def _parallel(
    ordered: bool,
    function: Callable,
    *iterables: Iterable,
    func_kws=None,
    num_cpus=None,
    **kwargs: Any
) -> Generator:
    """Returns a generator for a parallel map with a progress bar.

  Arguments:
      ordered(bool): True for an ordered map, false for an unordered map.
      function(Callable): The function to apply to each element of the given Iterables.
      iterables(Tuple[Iterable]): One or more Iterables containing the data to be mapped.

  Returns:
      A generator which will apply the function to each element of the given Iterables
      in parallel in order with a progress bar.
  """

    if func_kws is None:
        func_kws = {}

    if num_cpus is None:  # Determine num_cpus
        num_cpus = cpu_count()
    elif type(num_cpus) == float:
        num_cpus = int(round(num_cpus * cpu_count()))

    # Determine length of tqdm (equal to length of shortest iterable)
    length = min(len(iterable) for iterable in iterables if isinstance(iterable, Sized))

    map_type = "imap" if ordered else "uimap"
    pool = Pool(num_cpus)  # Create parallel generator
    map_func = getattr(pool, map_type)

    for item in tqdm(
        map_func(function, *iterables, (list(func_kws.values()),) * length),
        total=length,
        **kwargs
    ):
        yield item

    pool.clear()


def parallel_imap(function: Callable, *iterables: Iterable, **kwargs: Any) -> Generator:
    """Returns a generator for a parallel ordered map with a progress bar."""
    return _parallel(True, function, *iterables, **kwargs)


def parallel_map(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel ordered map with a progress bar."""
    return list(parallel_imap(function, *iterables, **kwargs))


def parallel_uimap(
    function: Callable, *iterables: Iterable, **kwargs: Any
) -> Generator:
    """Returns a generator for a parallel unordered map with a progress bar."""
    return _parallel(False, function, *iterables, **kwargs)


def parallel_umap(function: Callable, *iterables: Iterable, **kwargs: Any) -> List[Any]:
    """Performs a parallel unordered map with a progress bar."""
    return list(parallel_uimap(function, *iterables, **kwargs))


def sequential_imap(
    function: Callable, *iterables: Iterable, **kwargs: Any
) -> Generator:
    """Returns a generator for a sequential map with a progress bar."""
    return _sequential(function, *iterables, **kwargs)


def sequential_map(
    function: Callable, *iterables: Iterable, **kwargs: Any
) -> List[Any]:
    """Performs a sequential map with a progress bar."""
    return list(sequential_imap(function, *iterables, **kwargs))


if __name__ == "__main__":

    def asdasd():
        def add(a, b, *c):
            return a + b

        print(
            parallel_map(
                add, ["1", "2", "3"], ["a", "b", "c"], func_kws={"k": "sdafasd"}
            )
        )

    asdasd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import torch

__author__ = "Christian Heider Nielsen"

import numpy

__all__ = ["atari_initializer", "initialise_parameters"]

from draugr.torch_utilities.optimisation.parameters.initialisation.ortho_weight_init import (
    orthogonal_weights,
)


def atari_initializer(module: torch.nn.Module):
    """Parameter initializer for Atari models

  Initializes Linear, Conv2d, and LSTM weights."""
    classname = module.__class__.__name__

    if classname == "Linear":
        module.weight.data = orthogonal_weights(
            module.weight.data.size(), scale=numpy.sqrt(2.0)
        )
        module.bias.data.zero_()

    elif classname == "Conv2d":
        module.weight.data = orthogonal_weights(
            module.weight.data.size(), scale=numpy.sqrt(2.0)
        )
        module.bias.data.zero_()

    elif classname == "LSTM":
        for name, param in module.named_parameters():
            if "weight_ih" in name:
                param.data = orthogonal_weights(param.data.size(), scale=1.0)
            if "weight_hh" in name:
                param.data = orthogonal_weights(param.data.size(), scale=1.0)
            if "bias" in name:
                param.data.zero_()


def initialise_parameters(m: torch.nn.Module) -> None:
    """

  :param m:
  :type m:"""
    classname = m.__class__.__name__
    if classname.find("Linear") != -1:
        m.weight.data.normal_(0, 1)
        m.weight.data *= 1 / torch.sqrt(m.weight.data.pow(2).sum(1, keepdim=True))
        if m.bias is not None:
            m.bias.data.fill_(0)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
__author__ = "Christian Heider Nielsen"

__doc__ = r"""
"""

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .config import *
from .model import *
from .parameters import *
from .naming import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Sequence

import torch
from torch.utils.data import Dataset

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 09/10/2019
           """

__all__ = ["RandomDataset"]


class RandomDataset(Dataset):
    """"""

    def __init__(self, nd_size: Sequence, length: int):
        self.len = length
        self.data = torch.randn((length, *nd_size))

    def __getitem__(self, index):
        return self.data[index]

    def __len__(self):
        return self.len


if __name__ == "__main__":
    print(RandomDataset((5, 5), 10)[0])



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 30-12-2020
           """

if __name__ == "__main__":

    from draugr.tqdm_utilities import progress_bar

    for a in progress_bar(range(100)):
        print(a)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 11/05/2020
           """

from collections import OrderedDict
from itertools import tee

import torch
from draugr.torch_utilities.optimisation.parameters.freezing import freeze_parameters
from warg import AlsoDecorator

__all__ = [
    "TorchEvalSession",
    "TorchTrainSession",
    "TorchFrozenModelSession",
    "TorchTrainingSession",
]


class TorchEvalSession(AlsoDecorator):
    """
  # speed up evaluating after training finished
  """

    def __init__(self, model: torch.nn.Module, no_side_effect: bool = True):
        self.model = model
        self._no_side_effect = no_side_effect
        if no_side_effect:
            self.prev_state = model.training

    def __enter__(self):
        # self.model.eval()
        self.model.train(False)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            self.model.train(self.prev_state)
        else:
            self.model.train(True)


class TorchTrainSession(AlsoDecorator):
    """
  # speed up evaluating after training finished
  """

    def __init__(self, model: torch.nn.Module, no_side_effect: bool = True):
        self.model = model
        self._no_side_effect = no_side_effect
        if no_side_effect:
            self.prev_state = model.training

    def __enter__(self):
        self.model.train(True)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            self.model.train(self.prev_state)
        else:
            self.model.train(False)


TorchTrainingSession = TorchTrainSession


class TorchFrozenModelSession(AlsoDecorator):
    """"""

    def __init__(self, model: torch.nn.Module, no_side_effect: bool = True):
        self.model = model
        self._no_side_effect = no_side_effect
        self.params_1, self.params_2, self.params_3 = tee(model.parameters(True), 3)
        if no_side_effect:
            self.previous_states = [a.requires_grad for a in self.params_3]

    def __enter__(self):
        freeze_parameters(self.params_1, True)
        return True

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._no_side_effect:
            [p.requires_grad_(rg) for p, rg in zip(self.params_2, self.previous_states)]
        else:
            freeze_parameters(self.params_2, False)


if __name__ == "__main__":

    def main():

        a = torch.nn.Sequential(
            OrderedDict(l1=torch.nn.Linear(3, 5), l2=torch.nn.Linear(5, 2))
        )
        p_iter = iter(a.parameters(True))
        l1_w = next(p_iter)
        l1_bias = next(p_iter)
        l1_bias.requires_grad_(False)

        def initial():
            for p in a.parameters(True):
                print(p.requires_grad)

        @TorchFrozenModelSession(a)
        def frozen():
            for p in a.parameters(True):
                print(p.requires_grad)

        def frozen_session():
            with TorchFrozenModelSession(a):
                for p in a.parameters(True):
                    print(p.requires_grad)

        initial()
        print()
        frozen()
        print()
        initial()
        print()
        frozen_session()
        print()
        initial()

    main()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/03/2020
           """

from functools import partial
from typing import Callable, Iterable, Union

import cv2

__all__ = ["frame_generator"]

from draugr import identity


def frame_generator(
    video_stream: cv2.VideoCapture,
    coder: Union[None, Callable] = partial(cv2.cvtColor, code=cv2.COLOR_BGR2RGB),
) -> Iterable:
    """

  @param video_stream:
  @param coder:
  """
    if coder is None:
        coder = identity
    while video_stream.isOpened():
        success, frame = video_stream.read()
        if success:
            yield coder(frame)
        else:
            break


if __name__ == "__main__":

    def asd():
        from tqdm import tqdm

        for image in tqdm(frame_generator(cv2.VideoCapture(0), None)):
            cv2.namedWindow("window_name", cv2.WINDOW_NORMAL)
            cv2.imshow("window_name", image)

            if cv2.waitKey(1) == 27:
                break  # esc to quit

    asd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 21/02/2020
           """

import torch

__all__ = ["standardise"]


def minus_one_one_unnormalise(
    x: torch.tensor, low: float = 0, high: float = 1
) -> torch.tensor:
    act_k = (high - low) / 2.0
    act_b = (high + low) / 2.0
    return act_k * x + act_b


def minus_one_one_normalise(
    x: torch.tensor, low: float = 0, high: float = 1
) -> torch.tensor:
    act_k_inv = 2.0 / (high - low)
    act_b = (high + low) / 2.0
    return act_k_inv * (x - act_b)


def standardise(x: torch.tensor, eps: float = 1e-6) -> torch.tensor:
    """

  :param eps:
  :param x:
  :return:"""
    x -= x.mean()
    x /= x.std() + eps
    return x


if __name__ == "__main__":
    print(standardise(torch.ones(10)))
    print(standardise(torch.ones((10, 1))))
    print(standardise(torch.ones((1, 10))))

    print(standardise(torch.diag(torch.ones(3))))

    print(standardise(torch.ones((1, 10)) * torch.rand((1, 10))))

    print(standardise(torch.rand((1, 10))))

    print(minus_one_one_normalise(7, 0, 10))
    print(minus_one_one_unnormalise(0.4, 0, 10))
    print(minus_one_one_normalise(minus_one_one_unnormalise(3.4, 3, 4), 3, 4))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 25-01-2021
           """

from abc import abstractmethod

__all__ = ["Drawer"]


class Drawer(object):
    @abstractmethod
    def draw(self, *args, **kwargs):
        raise NotImplementedError

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 30-09-2020
           """

from draugr.windows_utilities import WindowsSettings


def test_get_dark_mode():
    ws = WindowsSettings()
    dark_mode = ws.get_dark_mode()
    print(dark_mode)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 03/08/2020
           """

__all__ = ["GtkSettings"]

import gi  # PyGObject lib

gi.require_version("Gtk", "3.0")
gi.require_version("Gdk", "3.0")

from warg import SingletonMeta


class GtkSettings(metaclass=SingletonMeta):
    """"""

    def __init__(self):
        from gi.repository import Gtk
        from gi.repository import Gdk

        # self.settings = Gtk.Settings.get_default()
        self.settings = Gtk.Settings.get_for_screen(
            Gdk.get_default_root_window().get_screen()
        )

    def __getitem__(self, item):
        return self.settings.get_property(item)

    def __setitem__(self, key, value):
        self.settings.set_property(key, value)

    @property
    def all_settings(self):
        """"""
        for i in self.settings.list_properties():  # getting all existing properties
            yield i

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        return 0  # self.settings


if __name__ == "__main__":

    def abvdfj():
        """"""
        a = GtkSettings()

        def ofgof():
            """"""
            for b in a.all_settings:
                print(b)

        ofgof()

        print(a["gtk-dialogs-use-header"])
        # print(a['gtk-dialogs-use-header1'])

    abvdfj()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 17/07/2020
           """

__all__ = ["recursive_flatten_numpy"]

from typing import Sequence

import numpy


def recursive_flatten_numpy(nested: Sequence) -> numpy.ndarray:
    return numpy.array(nested).ravel()


if __name__ == "__main__":

    print(recursive_flatten_numpy([[numpy.zeros((2, 2))], [numpy.zeros((2, 2))]]))
    print(recursive_flatten_numpy(numpy.zeros((2, 2))))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 07/07/2020
           """

import functools
from collections import OrderedDict

import torch
from torch import nn

__all__ = ["IntermediateLayerGetter"]


class IntermediateLayerGetter:
    """"""

    def __init__(
        self, model: torch.nn.Module, return_layers: list = None, keep_output=True
    ):
        """Wraps a Pytorch module to get intermediate values, eg for getting intermediate activations

    Arguments:
    model {nn.module} -- The Pytorch module to call
    return_layers {dict} -- Dictionary with the selected submodules
    to return the output (format: {[current_module_name]: [desired_output_name]},
    current_module_name can be a nested submodule, e.g. submodule1.submodule2.submodule3)

    Keyword Arguments:
    keep_output {bool} -- If True model_output contains the final model's output
    in the other case model_output is None (default: {True})

    Returns:
    (mid_outputs {OrderedDict}, model_output {any}) -- mid_outputs keys are
    your desired_output_name (s) and their values are the returned tensors
    of those submodules (OrderedDict([(desired_output_name,tensor(...)), ...).
    See keep_output argument for model_output description.
    In case a submodule is called more than one time, all it's outputs are
    stored in a list."""
        self._model = model
        if return_layers:
            self.return_layers = return_layers
        else:
            self.return_layers = model.parameters(recurse=True)
        self.keep_output = keep_output

    @staticmethod
    def reduce_getattr(obj, attr, *args):
        """
    # using wonder's beautiful simplification: https://stackoverflow.com/questions/31174295/getattr-and-setattr-on-nested-objects/31174427?noredirect=1#comment86638618_31174427

    :param obj:
    :type obj:
    :param attr:
    :type attr:
    :param args:
    :type args:
    :return:
    :rtype:"""

        def _getattr(obj, attr):
            return getattr(obj, attr, *args)

        return functools.reduce(_getattr, (obj, *attr.split(".")))

    def __call__(self, *args, **kwargs):
        ret = OrderedDict()
        handles = []
        for name, new_name in self.return_layers.items():
            layer = IntermediateLayerGetter.reduce_getattr(self._model, name)

            def hook(module, input, output, new_name=new_name):
                """

        :param module:
        :type module:
        :param input:
        :type input:
        :param output:
        :type output:
        :param new_name:
        :type new_name:"""
                if new_name in ret:
                    if type(ret[new_name]) is list:
                        ret[new_name].append(output)
                    else:
                        ret[new_name] = [ret[new_name], output]
                else:
                    ret[new_name] = output

            try:
                h = layer.register_forward_hook(hook)
            except AttributeError as e:
                raise AttributeError(f"Module {name} not found")
            handles.append(h)

        output = self._model(*args, **kwargs)

        for h in handles:
            h.remove()

        return ret, output


if __name__ == "__main__":

    def adsad():
        class Model(nn.Module):
            """"""

            def __init__(self):
                super().__init__()

                self.fc1 = nn.Linear(2, 2)
                self.fc2 = nn.Linear(2, 2)
                self.nested = nn.Sequential(
                    nn.Sequential(nn.Linear(2, 2), nn.Linear(2, 3)), nn.Linear(3, 1)
                )
                self.interaction_idty = (
                    nn.Identity()
                )  # Simple trick for operations not performed as modules

            def forward(self, x):
                """

        :param x:
        :type x:
        :return:
        :rtype:"""
                x1 = self.fc1(x)
                x2 = self.fc2(x)

                interaction = x1 * x2
                self.interaction_idty(interaction)
                return self.nested(interaction)

        model = Model()
        return_layers = {
            "fc2": "fc2",
            "nested.0.1": "nested",
            "interaction_idty": "interaction",
        }
        mid_getter = IntermediateLayerGetter(
            model, return_layers=return_layers, keep_output=True
        )
        mid_outputs, model_output = mid_getter(torch.randn(1, 2))

        print(model_output)
        print(mid_outputs)

    adsad()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04-01-2021
           """

import numpy

__all__ = ["DisplaySampler"]


class DisplaySampler(object):
    """
    # A class that will downsample the data and recompute when zoomed.
  """

    def __init__(self, x_data, y_data, verbose: bool = False):
        self._y_data = y_data
        self._x_data = x_data
        self.max_points = 50
        self.delta = x_data[-1] - x_data[0]
        self.verbose = verbose

    def downsample(self, xstart, xend):

        mask = (self._x_data > xstart) & (
            self._x_data < xend
        )  # get the points in the view range

        mask = numpy.convolve([1, 1], mask, mode="same").astype(
            bool
        )  # dilate the mask by one to catch the points just outside
        # of the view range to not truncate the line

        ratio = max(
            numpy.sum(mask) // self.max_points, 1
        )  # sort out how many points to drop

        x_data, y_data = (
            self._x_data[mask][::ratio],
            self._y_data[mask][::ratio],
        )  # mask data and downsample data

        if self.verbose:
            print(f"using {len(y_data)} of {numpy.sum(mask)} visible points")

        return x_data, y_data

    def update(self, ax):
        """Update the line"""
        limits = ax.viewLim
        if numpy.abs(limits.width - self.delta) > 1e-8:
            self.delta = limits.width
            self.line.set_data(*self.downsample(*limits.intervalx))
            ax.figure.canvas.draw_idle()


if __name__ == "__main__":

    def asdsad():
        from matplotlib import pyplot

        # Create a signal
        xdata = numpy.linspace(16, 365, (365 - 16) * 4)
        ydata = numpy.sin(2 * numpy.pi * xdata / 153) + numpy.cos(
            2 * numpy.pi * xdata / 127
        )

        sampler_display = DisplaySampler(xdata, ydata)

        fig, ax = pyplot.subplots()

        # Hook up the line
        (sampler_display.line,) = ax.plot(xdata, ydata, "o-")
        ax.set_autoscale_on(False)  # Otherwise, infinite loop

        # Connect for changing the view limits
        ax.callbacks.connect("xlim_changed", sampler_display.update)
        ax.set_xlim(16, 365)

    asdsad()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Iterable, Iterator, Tuple, Union

import numpy
import torch
from torch.utils.data.dataloader import DataLoader

from draugr.torch_utilities.datasets import NonSequentialDataset
from draugr.torch_utilities.tensors import to_tensor
from warg import passes_kws_to

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/10/2019
           """
__all__ = ["to_tensor_generator", "batch_generator_torch", "to_device_iterator"]


@passes_kws_to(to_tensor)
def to_tensor_generator(iterable: Iterable, preload_next: bool = False, **kwargs):
    """

:param iterable:
:param preload_next:
:param kwargs:
:return:"""
    if preload_next:
        iterable_iter = iter(iterable)
        current = to_tensor(next(iterable_iter), **kwargs)
        kwargs["non_blocking"] = True
        while current is not None:
            next_ = to_tensor(next(iterable_iter), **kwargs)
            yield current
            current = next_
    else:
        for a in iterable:
            yield to_tensor(a, **kwargs)
    return


def to_device_iterator(
    data_iterator: Iterator, device: Union[torch.device, str]
) -> Tuple:
    """

:param data_iterator:
:param device:
"""
    if isinstance(data_iterator, Iterable):
        data_iterator = iter(data_iterator)
    try:
        while True:
            yield (to_tensor(i, device=device) for i in next(data_iterator))
    except StopIteration:
        pass


@passes_kws_to(DataLoader)
def batch_generator_torch(
    sized: numpy.ndarray, mini_batches: int = 10, shuffle: bool = True, **kwargs
) -> DataLoader:
    """

:param sized:
:param mini_batches:
:param shuffle:
:param kwargs:
:return:
"""

    dataset = NonSequentialDataset(sized)
    return DataLoader(
        dataset, batch_size=len(dataset) // mini_batches, shuffle=shuffle, **kwargs
    )


if __name__ == "__main__":
    from torchvision.transforms import transforms
    import numpy
    from draugr.generators.recycling_generator import batched_recycle
    from draugr import inner_map

    def s():

        a = iter(numpy.random.sample((5, 5, 5)))
        for a in to_device_iterator(a, "cpu"):
            d, *_ = a
            print(d)
            print(type(d))

    def sdiaj():
        # a = numpy.random.sample((5, 5, 5))
        from draugr.torch_utilities.datasets import RandomDataset

        a = DataLoader(RandomDataset((10, 2), 100), batch_size=4)
        for _ in range(4):
            for i, a in enumerate(to_device_iterator(a, "cpu")):
                d, *_ = a
                print(d)
                print(type(d))

    def asijda():
        a_transform = transforms.Compose(
            [
                transforms.ToPILImage("RGB"),
                transforms.Resize(224),
                transforms.CenterCrop(224),
                transforms.RandomHorizontalFlip(),
                transforms.ToTensor(),
            ]
        )

        channels_in = 3
        # channels_out = 3

        # samples = 10
        device = "cuda"
        batches = 3
        batch_size = 32
        data_shape = (batches * batch_size, 256, 256, channels_in)

        generator = to_tensor_generator(
            inner_map(
                a_transform,
                batched_recycle(numpy.random.sample(data_shape), batch_size),
            ),
            device=device,
        )

        for i, a in enumerate(generator):
            print(a)
            break

    # s()
    sdiaj()
    # asijda()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import List, Union


def python_version_check(major=3, minor=6):
    import sys

    assert sys.version_info.major == major and sys.version_info.minor >= minor, (
        f"This project is utilises language features only present Python {major}.{minor} and greater. "
        f"You are running {sys.version_info}."
    )


python_version_check()

import pathlib
import re

from setuptools import find_packages, setup

with open(
    pathlib.Path(__file__).parent / "draugr" / "__init__.py", "r"
) as project_init_file:
    content = project_init_file.read()  # get strings from module
    version = re.search(r"__version__ = ['\"]([^'\"]*)['\"]", content, re.M).group(1)
    project_name = re.search(r"__project__ = ['\"]([^'\"]*)['\"]", content, re.M).group(
        1
    )
    author = re.search(r"__author__ = ['\"]([^'\"]*)['\"]", content, re.M).group(1)

__author__ = author


class DraugrPackage:
    @property
    def test_dependencies(self) -> list:
        path = pathlib.Path(__file__).parent
        requirements_tests = []
        with open(path / "requirements_tests.txt") as f:
            requirements = f.readlines()

            for requirement in requirements:
                requirements_tests.append(requirement.strip())

        return requirements_tests

    @property
    def setup_dependencies(self) -> list:
        path = pathlib.Path(__file__).parent
        requirements_setup = []
        with open(path / "requirements_setup.txt") as f:
            requirements = f.readlines()

            for requirement in requirements:
                requirements_setup.append(requirement.strip())

        return requirements_setup

    @property
    def package_name(self) -> str:
        return project_name

    @property
    def url(self) -> str:
        return "https://github.com/cnheider/draugr"

    @property
    def download_url(self) -> str:
        return self.url + "/releases"

    @property
    def readme_type(self) -> str:
        return "text/markdown"

    @property
    def packages(self) -> List[Union[bytes, str]]:
        return find_packages(
            exclude=[
                # 'Path/To/Exclude'
            ]
        )

    @property
    def author_name(self) -> str:
        return author

    @property
    def author_email(self) -> str:
        return "christian.heider@alexandra.dk"

    @property
    def maintainer_name(self) -> str:
        return self.author_name

    @property
    def maintainer_email(self) -> str:
        return self.author_email

    @property
    def package_data(self) -> dict:
        emds = [str(p) for p in pathlib.Path(__file__).parent.rglob(".md")]
        return {"draugr": [*emds]}

    @property
    def entry_points(self) -> dict:
        return {
            "console_scripts": [
                # "name_of_executable = module.with:function_to_execute"
                "draugr-darkmode-toggle = draugr.entry_points.toggle_darkmode:main",
                "draugr-tb = draugr.entry_points.tensorboard_entry_point:main",
                "draugr-cpu = draugr.entry_points.cpu_usage_entry_point:main",
                "draugr-vis = draugr.entry_points.visdom_entry_point:main",
            ]
        }

    @property
    def extras(self) -> dict:
        these_extras = {
            # 'ExtraName':['package-name; platform_system == "System(Linux,Windows)"'
        }

        path: pathlib.Path = pathlib.Path(__file__).parent

        for file in path.iterdir():
            if file.name.startswith("requirements_"):

                requirements_group = []
                with open(str(file.absolute())) as f:
                    requirements = f.readlines()

                    for requirement in requirements:
                        requirements_group.append(requirement.strip())

                group_name_ = "_".join(file.name.strip(".txt").split("_")[1:])

                these_extras[group_name_] = requirements_group

        all_dependencies = []

        for group_name in these_extras:
            all_dependencies += these_extras[group_name]
        these_extras["all"] = all_dependencies

        return these_extras

    @property
    def requirements(self) -> list:
        requirements_out = []
        with open("requirements.txt") as f:
            requirements = f.readlines()

            for requirement in requirements:
                requirements_out.append(requirement.strip())

        return requirements_out

    @property
    def description(self) -> str:
        return "A package for plotting directly in your terminal"

    @property
    def readme(self) -> str:
        with open("README.md") as f:
            return f.read()

    @property
    def changelog(self) -> str:
        with open("CHANGELOG.md") as f:
            return f.read()

    @property
    def keyword(self) -> str:
        with open("KEYWORDS.md") as f:
            return f.read()

    @property
    def license(self) -> str:
        return "Apache License, Version 2.0"

    @property
    def classifiers(self) -> List[str]:
        return [
            "Development Status :: 4 - Beta",
            "Environment :: Console",
            "Intended Audience :: End Users/Desktop",
            "Intended Audience :: Developers",
            "License :: OSI Approved :: Apache Software License",
            "Operating System :: MacOS :: MacOS X",
            "Operating System :: Microsoft :: Windows",
            "Operating System :: POSIX",
            "Operating System :: OS Independent",
            "Programming Language :: Python :: 3",
            "Natural Language :: English",
            # 'Topic :: Scientific/Engineering :: Artificial Intelligence'
            # 'Topic :: Software Development :: Bug Tracking',
        ]

    @property
    def version(self) -> str:
        return version


if __name__ == "__main__":

    pkg = DraugrPackage()

    setup(
        name=pkg.package_name,
        version=pkg.version,
        packages=pkg.packages,
        package_data=pkg.package_data,
        author=pkg.author_name,
        author_email=pkg.author_email,
        maintainer=pkg.maintainer_name,
        maintainer_email=pkg.maintainer_email,
        description=pkg.description,
        license=pkg.license,
        keywords=pkg.keyword,
        url=pkg.url,
        download_url=pkg.download_url,
        install_requires=pkg.requirements,
        extras_require=pkg.extras,
        setup_requires=pkg.setup_dependencies,
        entry_points=pkg.entry_points,
        classifiers=pkg.classifiers,
        long_description_content_type=pkg.readme_type,
        long_description=pkg.readme,
        tests_require=pkg.test_dependencies,
        include_package_data=True,
        python_requires=">=3.6",
    )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import torch

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/11/2019
           """

GLOBAL_DTYPE = None

__all__ = ["global_torch_dtype", "set_global_torch_dtype"]


def global_torch_dtype(
    override: torch.dtype = None, verbose: bool = False
) -> torch.dtype:
    """

  first time call stores to dtype for global reference, later call must manually override

  :param verbose:
  :type verbose:
  :param cuda_if_available:
  :type cuda_if_available:
  :param override:
  :type override:
  :return:
  :rtype:"""
    global GLOBAL_DTYPE

    if override is not None:
        GLOBAL_DTYPE = override
        set_global_torch_dtype(GLOBAL_DTYPE)
        if verbose:
            print(f"Overriding global torch device to {override}")
    elif GLOBAL_DTYPE is None:
        GLOBAL_DTYPE = torch.get_default_dtype()

    return GLOBAL_DTYPE


def set_global_torch_dtype(dtype: torch.dtype) -> None:
    global GLOBAL_DTYPE
    GLOBAL_DTYPE = dtype
    torch.set_default_dtype(GLOBAL_DTYPE)


if __name__ == "__main__":

    def stest_override():
        print(global_torch_dtype(verbose=True))
        print(global_torch_dtype(override=torch.double, verbose=True))
        print(global_torch_dtype(verbose=True))
        print(global_torch_dtype())
        print(global_torch_dtype())
        print(global_torch_dtype(override=torch.half))
        print(global_torch_dtype())

    stest_override()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 14/01/2020
           """

import pathlib

with open(pathlib.Path(__file__).parent / "README.md", "r") as this_init_file:
    __doc__ += this_init_file.read()

from .business import *
from .datetimes import *
from .debug import *
from .exceptions import *
from .manipulation import *
from .replication import *
from .sockets import *
from .styling import *
from .torch_like_channel_transformation import *
from .resources import *
from .platform_selection import *
from .strings import *
from .resources import *
from .powers import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import pathlib

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 23/09/2019
           """


def test_resolve_expand_path():
    # file_path = input("Enter another file path: ")
    file_path = pathlib.Path.home()
    model_save_path = pathlib.Path(file_path).expanduser().resolve()
    print(model_save_path)


if __name__ == "__main__":
    test_resolve_expand_path()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 30/03/2020
           """

from typing import Tuple

COLOR_RGB = Tuple[int, int, int]
COLOR_RGBA = Tuple[int, int, int, int]
COLOR_INT = Tuple[int, ...]

__all__ = ["RGB", "RGBA", "color_to_str", "color_from_str"]


def RGB(R: int, G: int, B: int) -> COLOR_RGB:
    """

  :param R:
  :type R:
  :param G:
  :type G:
  :param B:
  :type B:
  :return:
  :rtype:"""
    assert 0 <= R <= 255 and 0 <= G <= 255 and 0 <= B <= 255
    return R, G, B


def RGBA(R: int, G: int, B: int, A: int) -> COLOR_RGBA:
    """

  :param R:
  :type R:
  :param G:
  :type G:
  :param B:
  :type B:
  :param A:
  :type A:
  :return:
  :rtype:"""
    assert 0 <= R <= 255 and 0 <= G <= 255 and 0 <= B <= 255 and 0 <= A <= 255
    return R, G, B, A


def color_from_str(s: str, seperator: str = " ") -> COLOR_INT:
    """

  :param s:
  :type s:
  :param seperator:
  :type seperator:
  :return:
  :rtype:"""
    components = s.split(seperator)
    n = len(components)
    if n == 3:
        return RGB(*[int(c) for c in components])
    elif n == 4:
        return RGBA(*[int(c) for c in components])
    raise NotImplementedError("Color space not recognised")


def color_to_str(t: COLOR_INT, seperator: str = " ") -> str:
    """

  :param t:
  :type t:
  :param seperator:
  :type seperator:
  :return:
  :rtype:"""
    return seperator.join([str(c) for c in t])


if __name__ == "__main__":

    def main():
        a = RGB(1, 50, 100)
        b = color_to_str(a)
        c = color_from_str(b)
        print(c)

    main()



import multiprocessing
import platform
import typing

import torch

__all__ = ["system_info", "cuda_info"]


def system_info() -> str:
    """

  :return:
  :rtype:"""
    return "\n".join(
        [
            f"Python version: {platform.python_version()}",
            f"Python implementation: {platform.python_implementation()}",
            f"Python compiler: {platform.python_compiler()}",
            f"PyTorch version: {torch.__version__}",
            f"System: {platform.system() or 'Unable to determine'}",
            f"System version: {platform.release() or 'Unable to determine'}",
            f"Processor: {platform.processor() or 'Unable to determine'}",
            f"Number of CPUs: {multiprocessing.cpu_count()}",
        ]
    )


def cuda_info() -> str:
    """

  :return:
  :rtype:"""

    def _cuda_devices_formatting(
        info_function: typing.Callable,
        formatting_function: typing.Callable = None,
        mapping_function: typing.Callable = None,
    ):
        def _setup_default(function):
            return (lambda arg: arg) if function is None else function

        formatting_function = _setup_default(formatting_function)
        mapping_function = _setup_default(mapping_function)

        return " | ".join(
            mapping_function(
                [
                    formatting_function(info_function(i))
                    for i in range(torch.cuda.device_count())
                ]
            )
        )

    def _device_properties(attribute):
        return _cuda_devices_formatting(
            lambda i: getattr(torch.cuda.get_device_properties(i), attribute),
            mapping_function=lambda in_bytes: map(str, in_bytes),
        )

    cuda_cap = _cuda_devices_formatting(
        torch.cuda.get_device_capability,
        formatting_function=lambda capabilities: ".".join(map(str, capabilities)),
    )
    return "\n".join(
        [
            f"Available CUDA devices count: {torch.cuda.device_count()}",
            f"CUDA devices names: {_cuda_devices_formatting(torch.cuda.get_device_name)}",
            f"Major.Minor CUDA capabilities of devices: {cuda_cap}",
            f"Device total memory (bytes): {_device_properties('total_memory')}",
            f"Device multiprocessor count: {_device_properties('multi_processor_count')}",
        ]
    )


if __name__ == "__main__":
    print(system_info())
    print(cuda_info())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 31-10-2020
           """

from pathlib import Path

import torch
from pytorch_memlab import LineProfiler

from apppath import ensure_existence

if __name__ == "__main__":

    def inner():
        torch.nn.Linear(100, 100).cuda()

    def outer():
        linear = torch.nn.Linear(100, 100).cuda()
        linear2 = torch.nn.Linear(100, 100).cuda()
        inner()

    with LineProfiler(outer, inner) as prof:
        outer()

    with open(ensure_existence(Path("exclude")) / "test.html", "w") as f:
        f.write(prof.display()._repr_html_())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15/06/2020
           """

from typing import Iterable, Sequence, Tuple

from matplotlib import pyplot
from warg import AlsoDecorator, passes_kws_to

__all__ = ["FigureSession", "SubplotSession"]


class FigureSession(AlsoDecorator):
    @passes_kws_to(pyplot.figure)
    def __init__(self, **kws):
        self.fig = pyplot.figure(**kws)

    def __enter__(self) -> pyplot.Figure:
        return self.fig

    def __exit__(self, exc_type, exc_val, exc_tb):
        pyplot.cla()
        pyplot.close(self.fig)
        pyplot.clf()


class SubplotSession(AlsoDecorator):
    @passes_kws_to(pyplot.subplots)
    def __init__(self, **kws):
        self.fig, axs = pyplot.subplots(**kws)
        if not isinstance(axs, Iterable):
            axs = (axs,)
        self.axs = axs

    def __enter__(self) -> Tuple[pyplot.Figure, Sequence[pyplot.Axes]]:
        return self.fig, self.axs

    def __exit__(self, exc_type, exc_val, exc_tb):
        pyplot.cla()
        pyplot.close(self.fig)
        pyplot.clf()


if __name__ == "__main__":

    def deiajsd():
        for a in range(100):
            with SubplotSession() as a:
                fig, (ax1,) = a
                ax1.set_ylabel("test")

    deiajsd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

from typing import Iterator, Sequence, Tuple

from draugr.numpy_utilities import recursive_flatten_numpy
from draugr.drawers.mpl_drawers.mpldrawer import MplDrawer
from matplotlib import animation

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""

__all__ = ["DiscreteScrollPlot", "discrete_scroll_plot"]

from matplotlib import pyplot

import numpy
from warg import passes_kws_to


class DiscreteScrollPlot(MplDrawer):
    """
  Waterfall plot
  only supports a single trajectory at a time, do not supply parallel trajectories to draw method, will get truncated to num actions, effectively dropping actions for other envs than the first.
  """

    @passes_kws_to(MplDrawer.__init__)
    def __init__(
        self,
        num_actions: int,
        window_length: int = None,
        labels: Sequence = None,
        title: str = "",
        time_label: str = "Time",
        data_label: str = "Action Index",
        vertical: bool = True,
        reverse: bool = True,
        overwrite: bool = False,
        render: bool = True,
        figure_size: Tuple[int, int] = None,
        **kwargs
    ):

        super().__init__(render=render, figure_size=figure_size, **kwargs)
        if not render:
            return

        self._num_actions = num_actions

        if not window_length:
            window_length = num_actions * 20

        array = numpy.zeros((window_length, num_actions))

        self.vertical = vertical
        self.overwrite = overwrite
        self.reverse = reverse
        self.window_length = window_length

        if not figure_size:
            if vertical:
                self.fig = pyplot.figure(figsize=(window_length / 10, 2))
                extent = [-window_length, 0, 0, num_actions]
            else:
                self.fig = pyplot.figure(figsize=(2, window_length / 10))
                extent = [num_actions, 0, 0, -window_length]
        else:
            self.fig = pyplot.figure(figsize=figure_size)
            extent = [figure_size[0], 0, 0, -figure_size[1]]

        if vertical:
            array = array.T

        self.im = pyplot.imshow(
            array,
            cmap="gray",
            aspect="auto",
            interpolation="none",
            vmin=0.0,
            vmax=1.0,
            extent=extent,
        )

        b = numpy.arange(0.5, num_actions + 0.5, 1)
        if not labels:
            labels = numpy.arange(0, num_actions, 1)

        if vertical:
            pyplot.yticks(b, labels, rotation=45)
        else:
            pyplot.xticks(b, labels, rotation=45)

        if vertical:
            pyplot.xlabel(time_label)
            pyplot.ylabel(data_label)
        else:
            pyplot.xlabel(data_label)
            pyplot.ylabel(time_label)

        pyplot.title(title)
        pyplot.tight_layout()

    def _draw(self, data: Sequence[int]):
        """

    :param data:
    :return:"""

        array = self.im.get_array()

        data = recursive_flatten_numpy(data)
        # assert isinstance(data[0], int), f'Data was {data}'
        """
if not isinstance(data[0], int):
data = data[0]
"""

        data = data[: self._num_actions]

        if self.vertical:
            array = array.T

        if not self.overwrite:
            if not self.reverse:
                array = numpy.vstack((array[1:], data))
            else:
                array = numpy.vstack((data, array[:-1]))
        else:
            array[self.n % self.window_length] = data

        if self.vertical:
            array = array.T

        self.im.set_array(array)


def discrete_scroll_plot(
    vector_provider: Iterator,
    delta: float = 1 / 60,
    window_length: int = None,
    labels: Sequence = None,
    title: str = "",
    time_label: str = "Time",
    data_label: str = "Action Index",
    vertical: bool = True,
    reverse: bool = True,
    overwrite: bool = False,
):
    d = vector_provider.__next__()
    num_actions = len(d)
    if not window_length:
        window_length = num_actions * 20

    if not overwrite:
        array = numpy.zeros((window_length - 1, num_actions))
        if not reverse:
            array = numpy.vstack((array, d))
        else:
            array = numpy.vstack((d, array))
    else:
        array = numpy.zeros((window_length, num_actions))
        if not reverse:
            array[0] = d
        else:
            array[-1] = d

    def update_fig(n):
        data = vector_provider.__next__()
        array = im.get_array()

        if vertical:
            array = array.T

        if not overwrite:
            if not reverse:
                array = numpy.vstack((array[1:], data))
            else:
                array = numpy.vstack((data, array[:-1]))
        else:
            array[n % window_length] = data

        if vertical:
            array = array.T

        im.set_array(array)

        return im

    if vertical:
        fig = pyplot.figure(figsize=(window_length / 10, 2))
        extent = [-window_length, 0, 0, num_actions]
    else:
        fig = pyplot.figure(figsize=(2, window_length / 10))
        extent = [num_actions, 0, 0, -window_length]

    if vertical:
        array = array.T

    im = pyplot.imshow(
        array,
        cmap="gray",
        aspect="auto",
        interpolation="none",
        vmin=0.0,
        vmax=1.0,
        extent=extent,
    )

    b = numpy.arange(0.5, num_actions + 0.5, 1)
    if not labels:
        labels = numpy.arange(0, num_actions, 1)

    if vertical:
        pyplot.yticks(b, labels, rotation=45)
    else:
        pyplot.xticks(b, labels, rotation=45)

    if vertical:
        pyplot.xlabel(time_label)
        pyplot.ylabel(data_label)
    else:
        pyplot.xlabel(data_label)
        pyplot.ylabel(time_label)

    pyplot.title(title)
    pyplot.tight_layout()

    anim = animation.FuncAnimation(
        fig, update_fig, blit=False, fargs=(), interval=delta
    )

    return anim


if __name__ == "__main__":

    def siajdisajd():
        import queue
        import threading

        def ma():
            data = queue.Queue(100)

            class QueueGen:
                def __iter__(self):
                    return self

                def __next__(self):
                    return self.get()

                def __call__(self, *args, **kwargs):
                    return self.__next__()

                def add(self, a):
                    return data.put(a)

                def get(self):
                    return data.get()

            def get_sample(num_actions=3):
                a = numpy.zeros(num_actions)
                a[numpy.random.randint(0, num_actions)] = 1.0
                return a

            class MyDataFetchClass(threading.Thread):
                def __init__(self, data):

                    threading.Thread.__init__(self)

                    self._data = data

                def run(self):

                    while True:
                        self._data.add(get_sample())

            d = QueueGen()

            MyDataFetchClass(d).start()

            anim = discrete_scroll_plot(iter(d), labels=("a", "b", "c"))

            try:
                pyplot.show()
            except:
                print("Plot Closed")

        def asda():

            s = DiscreteScrollPlot(3, default_delta=None)
            for _ in range(100):
                s.draw(numpy.random.rand(3))

        asda()

    siajdisajd()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 04/03/2020
           """

from pathlib import Path
from typing import Sequence, Tuple, Union

import PIL.ImageDraw as ImageDraw
import PIL.ImageFont as ImageFont
import numpy
from PIL import Image
from draugr.opencv_utilities.bounding_boxes.colors import compute_color_for_labels
from draugr.opencv_utilities.opencv_draw import draw_masks
from draugr.python_utilities.colors import RGB

__all__ = ["draw_bounding_boxes"]


def draw_single_box(
    image: Image.Image,
    x_min: float,
    y_min: float,
    x_max: float,
    y_max: float,
    *,
    score_color: Tuple = RGB(0, 0, 0),
    display_str: str = None,
    font_type: ImageFont = ImageFont.load_default(),
    outline_color: Tuple = RGB(0, 255, 0),
    outline_width: int = 2,
    outline_alpha: float = 0.5,
    color_fill_score: bool = False,
) -> Image.Image:
    if not isinstance(image, ImageDraw.ImageDraw):
        assert isinstance(image, Image.Image)
        if image.mode != "RGBA":
            image = image.convert("RGBA")
        draw = ImageDraw.Draw(image, mode="RGBA")
    else:
        draw = image

    left, right, top, bottom = x_min, x_max, y_min, y_max
    alpha_color = outline_color + (int(255 * outline_alpha),)
    draw.rectangle(
        [(left, top), (right, bottom)],
        outline=outline_color,
        fill=alpha_color if color_fill_score else None,
        width=outline_width,
    )
    if display_str:
        if font_type is None or not isinstance(
            font_type,
            (ImageFont.FreeTypeFont, ImageFont.ImageFont, ImageFont.TransposedFont),
        ):
            font_type = ImageFont.load_default()
            print("Loading default (Better Than Nothing) font")

        text_bottom = bottom
        # Reverse list and print from bottom to top.
        text_width, text_height = font_type.getsize(display_str)
        margin = numpy.ceil(0.05 * text_height)
        draw.rectangle(
            xy=[
                (
                    left + outline_width,
                    text_bottom - text_height - 2 * margin - outline_width,
                ),
                (left + text_width + outline_width, text_bottom - outline_width),
            ],
            fill=alpha_color,
        )
        draw.text(
            (
                left + margin + outline_width,
                text_bottom - text_height - margin - outline_width,
            ),
            display_str,
            fill=score_color,
            font=font_type,
        )

    return image


def draw_bounding_boxes(
    image: Union[numpy.ndarray, Image.Image],
    boxes: numpy.ndarray,
    *,
    labels: numpy.ndarray = None,
    scores: numpy.ndarray = None,
    categories: Sequence[str] = None,
    outline_width: int = 2,
    outline_alpha: float = 0.5,
    score_color_fill: bool = False,
    score_font: ImageFont = ImageFont.load_default(),
    score_format: str = ": {:.2f}",
    mode="RGBA",
) -> numpy.ndarray:
    """Draw bounding boxes(labels, scores) on image
  Args:
  image: numpy array image, shape should be (height, width, channel)
  boxes: bboxes, shape should be (N, 4), and each row is (xmin, ymin, xmax, ymax), NOT NORMALISED!
  labels: labels, shape: (N, )
  scores: label scores, shape: (N, )
  class_name_map: list or dict, map class id to class name for visualization.
  width: box width
  alpha: text background alpha
  fill: fill box or not
  font: text font
  score_format: score format
  Returns:
  An image with information drawn on it."""
    boxes = numpy.array(boxes)
    num_boxes = boxes.shape[0]
    if isinstance(image, Image.Image):
        draw_image = image
    elif isinstance(image, numpy.ndarray):
        draw_image = Image.fromarray(image, mode)
    else:
        raise AttributeError(f"Unsupported images type {type(image)}")

    draw = ImageDraw.Draw(image, mode="RGBA")

    for i in range(num_boxes):
        display_str = ""
        color = (0, 255, 0)

        if labels is not None:
            ith_cat = int(labels[i])
            color = compute_color_for_labels(ith_cat)
            class_name = categories[ith_cat] if categories is not None else str(ith_cat)
            display_str += class_name

        if scores is not None:
            display_str += f"{score_format.format(scores[i])}"

        draw_image = draw_single_box(
            image=draw_image,
            x_min=boxes[i, 0],
            y_min=boxes[i, 1],
            x_max=boxes[i, 2],
            y_max=boxes[i, 3],
            outline_color=color,
            outline_width=outline_width,
            outline_alpha=outline_alpha,
            display_str=display_str,
            font_type=score_font,
            color_fill_score=score_color_fill,
        )

    return numpy.array(draw_image, dtype=numpy.uint8)


if __name__ == "__main__":

    def a():
        from matplotlib import pyplot
        import pickle
        from neodroidvision.data.datasets.supervised.detection.coco import COCODataset

        name = "000000308476"
        data_root = Path.home() / "Data" / "Coco"
        with open(str(data_root / f"{name}.pickle"), "rb") as f:
            data = pickle.load(f)

        img = Image.open(str(data_root / f"{name}.jpg"))
        img = draw_masks(img, data.masks, data.labels)
        img = draw_bounding_boxes(
            img,
            boxes=data.boxes,
            labels=data.labels,
            scores=data.scores,
            categories=COCODataset.categories,
            score_format=": {:.2f}",
        )
        pyplot.imshow(img)
        pyplot.show()

    def b():
        from matplotlib import pyplot

        data_root = Path.home() / "Data" / "PennFudanPed" / "PNGImages"

        img = Image.open(str(data_root / f"FudanPed00001.png"))
        img_width, _ = img.size
        img = draw_bounding_boxes(
            img,
            boxes=(
                (0 * img_width, 0 * img_width, 0.2 * img_width, 0.2 * img_width),
                (0.3 * img_width, 0.3 * img_width, 0.5 * img_width, 0.5 * img_width),
                (0.1 * img_width, 0.1 * img_width, 0.7 * img_width, 0.7 * img_width),
            ),
            labels=(0, 55, 78),
            scores=(0.2, 1.0, 0.4),
            categories=(*[f"cat{i}" for i in range(99)],),
            score_format=": {:.2f}",
            outline_width=5,
            score_color_fill=False,
        )
        pyplot.imshow(img)
        pyplot.show()

    b()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 01/03/2020
           """

import torch

__all__ = ["warmup_lr_scheduler"]


def warmup_lr_scheduler(optimizer, warmup_iters, warmup_factor):
    """

  :param optimizer:
  :type optimizer:
  :param warmup_iters:
  :type warmup_iters:
  :param warmup_factor:
  :type warmup_factor:
  :return:
  :rtype:"""

    def f(x):
        """

    :param x:
    :type x:
    :return:
    :rtype:"""
        if x >= warmup_iters:
            return 1
        alpha = float(x) / warmup_iters
        return warmup_factor * (1 - alpha) + alpha

    return torch.optim.lr_scheduler.LambdaLR(optimizer, f)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 15-10-2020
           """

import pyperf

runner = pyperf.Runner()
runner.timeit(
    name="sort a sorted list",
    stmt="sorted(s, key=f)",
    setup="f = lambda x: x; s = list(range(1000))",
)



# To add a new cell, type '# %%'
# To add a new markdown cell, type '# %% [markdown]'
# %%
from IPython import get_ipython

# %% [markdown]
# ##### Copyright 2019 The TensorFlow Authors.

# %%
# @title Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# %% [markdown]
# # Hyperparameter Tuning with the HParams Dashboard
#
# <table class="tfo-notebook-buttons" align="left">
#   <td>
#     <a target="_blank" href="https://www.tensorflow.org/tensorboard/hyperparameter_tuning_with_hparams"><img src="https://www.tensorflow.org/images/tf_logo_32px.png" />View on TensorFlow.org</a>
#   </td>
#   <td>
#     <a target="_blank" href="https://colab.research.google.com/github/tensorflow/tensorboard/blob/master/docs/hyperparameter_tuning_with_hparams.ipynb"><img src="https://www.tensorflow.org/images/colab_logo_32px.png" />Run in Google Colab</a>
#   </td>
#   <td>
#     <a target="_blank" href="https://github.com/tensorflow/tensorboard/blob/master/docs/hyperparameter_tuning_with_hparams.ipynb"><img src="https://www.tensorflow.org/images/GitHub-Mark-32px.png" />View source on GitHub</a>
#   </td>
# </table>
# %% [markdown]
# When building machine learning models, you need to choose various [hyperparameters](https://en.wikipedia.org/wiki/Hyperparameter_(machine_learning)), such as the dropout rate in a layer or the learning rate. These decisions impact model metrics, such as accuracy. Therefore, an important step in the machine learning workflow is to identify the best hyperparameters for your problem, which often involves experimentation. This process is known as "Hyperparameter Optimization" or "Hyperparameter Tuning".
#
# The HParams dashboard in TensorBoard provides several tools to help with this process of identifying the best experiment or most promising sets of hyperparameters.
#
# This tutorial will focus on the following steps:
#
# 1. Experiment setup and HParams summary
# 2. Adapt TensorFlow runs to log hyperparameters and metrics
# 3. Start runs and log them all under one parent directory
# 4. Visualize the results in TensorBoard's HParams dashboard
#
# Note: The HParams summary APIs and dashboard UI are in a preview stage and will change over time.
#
# Start by installing TF 2.0 and loading the TensorBoard notebook extension:

# %%
# Load the TensorBoard notebook extension
get_ipython().run_line_magic("load_ext", "tensorboard")

# %%
# Clear any logs from previous runs
get_ipython().system("rm -rf ./logs/ ")

# %% [markdown]
# Import TensorFlow and the TensorBoard HParams plugin:

# %%
import tensorflow as tf
from tensorboard.plugins.hparams import api as hp

# %% [markdown]
# Download the [FashionMNIST](https://github.com/zalandoresearch/fashion-mnist) dataset and scale it:

# %%
fashion_mnist = tf.keras.datasets.fashion_mnist

(x_train, y_train), (x_test, y_test) = fashion_mnist.load_data()
x_train, x_test = x_train / 255.0, x_test / 255.0

# %% [markdown]
# ## 1. Experiment setup and the HParams experiment summary
#
# Experiment with three hyperparameters in the model:
#
# 1. Number of units in the first dense layer
# 2. Dropout rate in the dropout layer
# 3. Optimizer
#
# List the values to try, and log an experiment configuration to TensorBoard. This step is optional: you can provide domain information to enable more precise filtering of hyperparameters in the UI, and you can specify which metrics should be displayed.

# %%
HP_NUM_UNITS = hp.HParam("num_units", hp.Discrete([16, 32]))
HP_DROPOUT = hp.HParam("dropout", hp.RealInterval(0.1, 0.2))
HP_OPTIMIZER = hp.HParam("optimizer", hp.Discrete(["adam", "sgd"]))

METRIC_ACCURACY = "accuracy"

with tf.summary.create_file_writer("logs/hparam_tuning").as_default():
    hp.hparams_config(
        hparams=[HP_NUM_UNITS, HP_DROPOUT, HP_OPTIMIZER],
        metrics=[hp.Metric(METRIC_ACCURACY, display_name="Accuracy")],
    )


# %% [markdown]
# If you choose to skip this step, you can use a string literal wherever you would otherwise use an `HParam` value: e.g., `hparams['dropout']` instead of `hparams[HP_DROPOUT]`.
# %% [markdown]
# ## 2. Adapt TensorFlow runs to log hyperparameters and metrics
#
# The model will be quite simple: two dense layers with a dropout layer between them. The training code will look familiar, although the hyperparameters are no longer hardcoded. Instead, the hyperparameters are provided in an `hparams` dictionary and used throughout the training function:

# %%
def train_test_model(hparams):
    model = tf.keras.models.Sequential(
        [
            tf.keras.layers.Flatten(),
            tf.keras.layers.Dense(hparams[HP_NUM_UNITS], activation=tf.nn.relu),
            tf.keras.layers.Dropout(hparams[HP_DROPOUT]),
            tf.keras.layers.Dense(10, activation=tf.nn.softmax),
        ]
    )
    model.compile(
        optimizer=hparams[HP_OPTIMIZER],
        loss="sparse_categorical_crossentropy",
        metrics=["accuracy"],
    )

    model.fit(
        x_train, y_train, epochs=1
    )  # Run with 1 epoch to speed things up for demo purposes
    _, accuracy = model.evaluate(x_test, y_test)
    return accuracy


# %% [markdown]
# For each run, log an hparams summary with the hyperparameters and final accuracy:

# %%
def run(run_dir, hparams):
    with tf.summary.create_file_writer(run_dir).as_default():
        hp.hparams(hparams)  # record the values used in this trial
        accuracy = train_test_model(hparams)
        tf.summary.scalar(METRIC_ACCURACY, accuracy, step=1)


# %% [markdown]
# When training Keras models, you can use callbacks instead of writing these directly:
#
# ```python
# model.fit(
#     ...,
#     callbacks=[
#         tf.keras.callbacks.TensorBoard(logdir),  # log metrics
#         hp.KerasCallback(logdir, hparams),  # log hparams
#     ],
# )
# ```
# %% [markdown]
# ## 3. Start runs and log them all under one parent directory
#
# You can now try multiple experiments, training each one with a different set of hyperparameters.
#
# For simplicity, use a grid search: try all combinations of the discrete parameters and just the lower and upper bounds of the real-valued parameter. For more complex scenarios, it might be more effective to choose each hyperparameter value randomly (this is called a random search). There are more advanced methods that can be used.
#
# Run a few experiments, which will take a few minutes:

# %%
session_num = 0

for num_units in HP_NUM_UNITS.domain.values:
    for dropout_rate in (HP_DROPOUT.domain.min_value, HP_DROPOUT.domain.max_value):
        for optimizer in HP_OPTIMIZER.domain.values:
            hparams = {
                HP_NUM_UNITS: num_units,
                HP_DROPOUT: dropout_rate,
                HP_OPTIMIZER: optimizer,
            }
            run_name = "run-%d" % session_num
            print("--- Starting trial: %s" % run_name)
            print({h.name: hparams[h] for h in hparams})
            run("logs/hparam_tuning/" + run_name, hparams)
            session_num += 1

# %% [markdown]
# ## 4. Visualize the results in TensorBoard's HParams plugin
# %% [markdown]
# The HParams dashboard can now be opened. Start TensorBoard and click on "HParams" at the top.

# %%
get_ipython().run_line_magic("tensorboard", "--logdir logs/hparam_tuning")

# %% [markdown]
# <img class="tfo-display-only-on-site" src="images/hparams_table.png?raw=1"/>
# %% [markdown]
# The left pane of the dashboard provides filtering capabilities that are active across all the views in the HParams dashboard:
#
# - Filter which hyperparameters/metrics are shown in the dashboard
# - Filter which hyperparameter/metrics values are shown in the dashboard
# - Filter on run status (running, success, ...)
# - Sort by hyperparameter/metric in the table view
# - Number of session groups to show (useful for performance when there are many experiments)
#
# %% [markdown]
# The HParams dashboard has three different views, with various useful information:
#
# * The **Table View** lists the runs, their hyperparameters, and their metrics.
# * The **Parallel Coordinates View** shows each run as a line going through an axis for each hyperparemeter and metric. Click and drag the mouse on any axis to mark a region which will highlight only the runs that pass through it. This can be useful for identifying which groups of hyperparameters are most important. The axes themselves can be re-ordered by dragging them.
# * The **Scatter Plot View** shows plots comparing each hyperparameter/metric with each metric. This can help identify correlations. Click and drag to select a region in a specific plot and highlight those sessions across the other plots.
#
# A table row, a parallel coordinates line, and a scatter plot market can be clicked to see a plot of the metrics as a function of training steps for that session (although in this tutorial only one step is used for each run).
# %% [markdown]
# To further explore the capabilities of the HParams dashboard, download a set of pregenerated logs with more experiments:

# %%
get_ipython().run_cell_magic(
    "bash",
    "",
    "wget -q 'https://storage.googleapis.com/download.tensorflow.org/tensorboard/hparams_demo_logs.zip'\nunzip -q hparams_demo_logs.zip -d logs/hparam_demo",
)

# %% [markdown]
# View these logs in TensorBoard:

# %%
get_ipython().run_line_magic("tensorboard", "--logdir logs/hparam_demo")

# %% [markdown]
# <img class="tfo-display-only-on-site" src="images/hparams_parallel_coordinates.png?raw=1"/>
# %% [markdown]
# You can try out the different views in the HParams dashboard.
#
# For example, by going to the parallel coordinates view and clicking and dragging on the accuracy axis, you can select the runs with the highest accuracy. As these runs pass through 'adam' in the optimizer axis, you can conclude that 'adam' performed better than 'sgd' on these experiments.



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

            Small utilities to keep track of cores dedicated to workers threads

           Created on 07-12-2020
           """

import os

from warg import AlsoDecorator, passes_kws_to

CORE_COUNT: int = os.cpu_count()
IN_USE_BY_THIS_PROCESS: int = 0

__all__ = [
    "request_worker_cores",
    "reset_worker_tracker",
    "release_worker_cores",
    "worker_cores_available",
    "worker_cores_in_use",
    "WorkerSession",
]


def request_worker_cores(
    percentage: float, *, of_remaining: bool = False, verbose: bool = False
) -> int:
    """
global_pin_memory
:param percentage:
:param of_remaining:
:param verbose:
:return:
"""
    global IN_USE_BY_THIS_PROCESS

    if IN_USE_BY_THIS_PROCESS >= CORE_COUNT:
        print(
            f"WARNING! (IN_USE_BY_THIS_PROCESS/CORES_AVAILABLE) {IN_USE_BY_THIS_PROCESS}/{CORE_COUNT}"
        )
        return 1

    if of_remaining:
        cores = round((CORE_COUNT - IN_USE_BY_THIS_PROCESS) * percentage)
    else:
        cores = round(CORE_COUNT * percentage)

    if verbose:
        print(f"reserving {cores} workers")

    IN_USE_BY_THIS_PROCESS += cores

    return cores


def release_worker_cores(num: int) -> int:
    """

:param num:
:return:
"""
    global IN_USE_BY_THIS_PROCESS
    res = max(IN_USE_BY_THIS_PROCESS - num, 0)
    IN_USE_BY_THIS_PROCESS = res
    return res


def core_count() -> int:
    """

:return:
"""
    return CORE_COUNT


def worker_cores_available() -> int:
    """
maybe negative if over allocated
:return:
"""
    return CORE_COUNT - IN_USE_BY_THIS_PROCESS


def worker_cores_in_use() -> int:
    """

:return:
"""
    return IN_USE_BY_THIS_PROCESS


def reset_worker_tracker() -> None:
    """

:return:
"""
    global IN_USE_BY_THIS_PROCESS
    IN_USE_BY_THIS_PROCESS = 0


class WorkerSession(AlsoDecorator):
    """
# speed up evaluating after training finished
"""

    @passes_kws_to(request_worker_cores)
    def __init__(self, percentage, **kwargs):
        self.percentage = percentage
        self.kws = kwargs

    def __enter__(self):
        self.num = request_worker_cores(self.percentage, **self.kws)
        return self.num

    def __exit__(self, exc_type, exc_val, exc_tb):
        release_worker_cores(self.num)


if __name__ == "__main__":
    print(worker_cores_available())
    print(worker_cores_in_use())
    print(request_worker_cores(0.5))
    print(worker_cores_available())
    print(worker_cores_in_use())
    release_worker_cores(round(core_count() * 0.5))
    print(worker_cores_in_use())
    print()
    print(worker_cores_available())
    print()
    with WorkerSession(0.33) as num_cores:
        print(num_cores)
        print(worker_cores_in_use())
        print(worker_cores_available())
    print(worker_cores_in_use())
    print()
    with WorkerSession(0.25) as num_cores:
        with WorkerSession(0.25) as num_cores_inner:
            print(num_cores)
            print(num_cores_inner)
            print(worker_cores_in_use())
            print(worker_cores_available())
    print(worker_cores_in_use())



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from typing import Union

import numpy
import torch
from PIL import Image
from draugr.metrics import MetricCollection
from draugr.writers.mixins import ImageWriterMixin
from draugr.writers.terminal.terminal_image_renderer import (
    render_image,
    terminalise_image,
)
from draugr.writers.terminal.terminal_plot import (
    styled_terminal_plot_stats_shared_x,
    terminal_plot,
)
from draugr.writers.writer import Writer
from tqdm import tqdm

__author__ = "Christian Heider Nielsen"
__doc__ = """
Created on 27/04/2019

@author: cnheider
"""
__all__ = ["TerminalPlotWriter"]


class TerminalPlotWriter(Writer, ImageWriterMixin):
    """"""

    def image(
        self,
        tag: str,
        data: Union[numpy.ndarray, torch.Tensor, Image.Image],
        step,
        *,
        dataformats: str = "NCHW",
        **kwargs,
    ) -> None:
        self.E.write(terminalise_image(render_image(data, scale=(28, 28))))

    def _open(self):
        self.E = tqdm()
        self.values = MetricCollection()
        return self

    def _close(self, exc_type=None, exc_val=None, exc_tb=None):
        return self.E.close()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self._stats_tag = "stats"

    def stats(self, value: MetricCollection, step_i: int = None):
        """

    :type value: object"""
        if step_i:
            if self.filter(self._stats_tag):
                self._stats(value, self._counter[self._stats_tag])
            self._counter[self._stats_tag] = step_i
        else:
            if self.filter(self._stats_tag):
                self._stats(value, self._counter[self._stats_tag])
            self._counter[self._stats_tag] += 1

    def _stats(self, stats: MetricCollection, step_i):
        styled_terminal_plot_stats_shared_x(stats, printer=self.E.write)
        self.E.set_description(
            f"Epi: {step_i}, "
            f"Sig: {stats.signal.running_value[-1]:.3f}, "
            f"Dur: {stats.duration.running_value[-1]:.1f}, "
            f"TD Err: {stats.td_error.running_value[-1]:.3f}, "
            f"Eps: {stats.epsilon.running_value[-1]:.3f}"
        )

    def _scalar(self, tag: str, value: float, step: int):
        self.values[tag] = value
        # styled_terminal_plot_stats_shared_x(self.values, printer=self.E.write)
        terminal_plot([value], printer=self.E.write)
        self.E.set_description(f"Tag:{tag} Val:{value} Step:{step}")


if __name__ == "__main__":

    with TerminalPlotWriter() as w:
        w.scalar("What", 4)
        w.image("bro", numpy.random.randint(0, 255, (28, 28, 3)), 0)

    '''
def train_episodically_old(self,
           env,
           test_env,
           *,
           rollouts=2000,
           render=False,
           render_frequency=100,
           stat_frequency=10,
           ):

E = range(1, rollouts)
E = tqdm(E, f"Episode: {1}", leave=False, disable=not render)

stats = draugr.StatisticCollection(stats=("signal", "duration", "entropy"))

for episode_i in E:
initial_state = env.reset()

if episode_i % stat_frequency == 0:
draugr.styled_terminal_plot_stats_shared_x(stats, printer=E.write)

E.set_description(
f"Epi: {episode_i}, "
f"Sig: {stats.signal.running_value[-1]:.3f}, "
f"Dur: {stats.duration.running_value[-1]:.1f}"
)

if render and episode_i % render_frequency == 0:
signal, dur, entropy, *extras = self.rollout(
initial_state, env, render=render
)
else:
signal, dur, entropy, *extras = self.rollout(initial_state, env)

stats.duration.append(dur)
stats.signal.append(signal)
stats.entropy.append(entropy)

if self.end_training:
break

return NOD(model=self._distribution_parameter_regressor, stats=stats)


def train_episodically_old(self,
           _environment,
           *,
           rollouts=10000,
           render=False,
           render_frequency=100,
           stat_frequency=100,
           **kwargs,
           ):
"""
:param _environment:
:type _environment:,0
:param rollouts:
:type rollouts:
:param render:
:type render:
:param render_frequency:
:type render_frequency:
:param stat_frequency:
:type stat_frequency:
:return:
:rtype:
"""

stats = draugr.StatisticCollection(
stats=("signal", "duration", "td_error", "epsilon")
)

E = range(1, rollouts)
E = tqdm(E, leave=False, disable=not render)

for episode_i in E:
initial_state = _environment.reset()

if episode_i % stat_frequency == 0:
draugr.styled_terminal_plot_stats_shared_x(stats, printer=E.write)
E.set_description(
f"Epi: {episode_i}, "
f"Sig: {stats.signal.running_value[-1]:.3f}, "
f"Dur: {stats.duration.running_value[-1]:.1f}, "
f"TD Err: {stats.td_error.running_value[-1]:.3f}, "
f"Eps: {stats.epsilon.running_value[-1]:.3f}"
)

if render and episode_i % render_frequency == 0:
signal, dur, td_error, *extras = self.rollout(
initial_state, _environment, render=render
)
else:
signal, dur, td_error, *extras = self.rollout(
initial_state, _environment
)

stats.append(signal, dur, td_error, self._current_eps_threshold)

if self.end_training:
break

return NOD(model=self._value_model, stats=stats)
'''



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20-01-2021
           """

from mpl_toolkits.mplot3d import Axes3D
from matplotlib import cm, pyplot
import numpy

__all__ = ["plot_gradients_3d"]


def plot_gradients_3d() -> pyplot.Figure:
    fig = pyplot.figure()
    ax = fig.gca(projection="3d")
    X = numpy.arange(-5, 5, 0.25)
    Y = numpy.arange(-5, 5, 0.25)
    X, Y = numpy.meshgrid(X, Y)
    R = numpy.sqrt(X ** 2 + Y ** 2)
    Z = numpy.sin(R)
    Gx, Gy = numpy.gradient(Z)  # gradients with respect to x and y
    G = (Gx ** 2 + Gy ** 2) ** 0.5  # gradient magnitude
    N = G / G.max()  # normalize 0..1
    surf = ax.plot_surface(
        X,
        Y,
        Z,
        rstride=1,
        cstride=1,
        facecolors=cm.jet(N),
        linewidth=0,
        antialiased=False,
        shade=False,
    )
    return fig


if __name__ == "__main__":
    plot_gradients_3d()
    pyplot.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 9/14/19
           """

from .terminal_plot import *
from .terminal_plot_writer import *
from .console_writer import *



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
import cv2
import numpy

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 10/10/2019
           """

__all__ = ["draw_cube", "draw_axis", "cube_3d_matrix"]

from warg import Number


def draw_axis(
    img,
    corners,
    rotation_vectors,
    translation_vectors,
    camera_matrix,
    dist_coef,
    size=1,
):
    """

  :param img:
  :type img:
  :param corners:
  :type corners:
  :param rotation_vectors:
  :type rotation_vectors:
  :param translation_vectors:
  :type translation_vectors:
  :param camera_matrix:
  :type camera_matrix:
  :param dist_coef:
  :type dist_coef:
  :param size:
  :type size:
  :return:
  :rtype:"""
    axis_size: numpy.ndarray = numpy.eye(3, 3) @ numpy.diag([size, size, -size])
    axis_size = axis_size.astype(numpy.float32)

    img_pts, jac = cv2.projectPoints(
        axis_size, rotation_vectors, translation_vectors, camera_matrix, dist_coef
    )  # project 3D points to image plane

    corner = tuple(corners[0].ravel())
    img = cv2.line(img, corner, tuple(img_pts[0].ravel()), (255, 0, 0), 5)
    img = cv2.line(img, corner, tuple(img_pts[1].ravel()), (0, 255, 0), 5)
    img = cv2.line(img, corner, tuple(img_pts[2].ravel()), (0, 0, 255), 5)
    return img


def cube_3d_matrix():
    """

  :return:
  :rtype:"""
    return [
        [0, 0, 0],
        [0, 1, 0],
        [1, 1, 0],
        [1, 0, 0],
        [0, 0, -1],
        [0, 1, -1],
        [1, 1, -1],
        [1, 0, -1],
    ]


def draw_cube(
    img,
    rotation_vectors,
    translation_vectors,
    camera_matrix,
    dist_coef,
    size: Number = 6,
):
    """

  :param img:
  :type img:
  :param rotation_vectors:
  :type rotation_vectors:
  :param translation_vectors:
  :type translation_vectors:
  :param camera_matrix:
  :type camera_matrix:
  :param dist_coef:
  :type dist_coef:
  :param size:
  :type size:
  :return:
  :rtype:"""
    cube_size = numpy.float32(cube_3d_matrix()) * size

    img_pts, jac2 = cv2.projectPoints(
        cube_size, rotation_vectors, translation_vectors, camera_matrix, dist_coef
    )

    img_pts = numpy.int32(img_pts).reshape(-1, 2)

    # draw ground floor in green
    img = cv2.drawContours(img, [img_pts[:4]], -1, (0, 255, 0), -3)

    # draw pillars in blue color
    for i, j in zip(range(4), range(4, 8)):
        img = cv2.line(img, tuple(img_pts[i]), tuple(img_pts[j]), (255, 0, 0), 3)

    # draw top layer in red color
    img = cv2.drawContours(img, [img_pts[4:]], -1, (0, 0, 255), 3)

    return img


if __name__ == "__main__":
    draw_cube()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 28/07/2020
           """

__all__ = ["orthogonal_reg", "RegLossWrapper"]

import torch
from torch import nn


class RegLossWrapper(torch.nn.Module):
    """"""

    def __init__(self, loss, model: torch.nn.Module, factor: float = 0.0005):
        super().__init__()
        self.loss = loss
        self.l1_crit = torch.nn.L1Loss()
        self.a = torch.zeros(1)
        self.factor = factor
        self.params = []
        for name, param in model.named_parameters():
            if "bias" not in name:
                self.params.append(param)

    def forward(self, *loss, **kwargs) -> torch.Tensor:
        return self.loss(*loss) + self.factor * sum(
            [self.l1_crit(p, self.a) for p in self.params]
        )


def orthogonal_reg(model, reg=1e-6):
    with torch.enable_grad():
        orth_loss = torch.zeros(1)
        for name, param in model.named_parameters():
            if "bias" not in name:
                param_flat = param.reshape(param.shape[0], -1)
                sym = torch.mm(param_flat, torch.t(param_flat))
                sym -= torch.eye(param_flat.shape[0])
                orth_loss += reg * sym.abs().sum()


if __name__ == "__main__":

    def bb():
        from draugr.torch_utilities.optimisation.parameters.initialisation import (
            normal_init_weights,
        )

        input = torch.randn(3, 5, requires_grad=True)
        model = torch.nn.Linear(5, 5)
        normal_init_weights(model)
        target = torch.empty(3, dtype=torch.long).random_(5)
        loss_fn = RegLossWrapper(nn.CrossEntropyLoss(), model)

        def a(m):
            loss = loss_fn(m(input), target)
            print(loss)
            loss.backward()

        a(model)
        a(model)
        normal_init_weights(model, 1.0)
        a(model)

    bb()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from draugr.writers import terminal_plot

__author__ = "Christian Heider Nielsen"

import numpy


def test_terminal_plot():
    terminal_plot(numpy.tile(range(9), 4), plot_character="o")
    assert True


if __name__ == "__main__":
    test_terminal_plot()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
from torch import nn

__author__ = "Christian Heider Nielsen"
__doc__ = r"""

           Created on 20/10/2019
           """

__all__ = ["normal_init_weights"]


def normal_init_weights(m, mean: float = 0.0, std: float = 0.1) -> None:
    """

    :param mean:
    :param std:
  :param m:
  :type m:"""
    if isinstance(m, nn.Linear):
        nn.init.normal_(m.weight, mean=mean, std=std)
        nn.init.constant_(m.bias, std)
